/*
 * Copyright (c) 2016, Alliance for Open Media. All rights reserved
 *
 * This source code is subject to the terms of the BSD 2 Clause License and
 * the Alliance for Open Media Patent License 1.0. If the BSD 2 Clause License
 * was not distributed with this source code in the LICENSE file, you can
 * obtain it at www.aomedia.org/license/software. If the Alliance for Open
 * Media Patent License 1.0 was not distributed with this source code in the
 * PATENTS file, you can obtain it at www.aomedia.org/license/patent.
 */

#include "av1/common/entropy.h"
#include "av1/common/blockd.h"
#include "av1/common/onyxc_int.h"
#include "av1/common/entropymode.h"
#include "aom_mem/aom_mem.h"
#include "aom/aom_integer.h"

// Unconstrained Node Tree
/* clang-format off */
const aom_tree_index av1_coef_con_tree[TREE_SIZE(ENTROPY_TOKENS)] = {
  2, 6,                                // 0 = LOW_VAL
  -TWO_TOKEN, 4,                       // 1 = TWO
  -THREE_TOKEN, -FOUR_TOKEN,           // 2 = THREE
  8, 10,                               // 3 = HIGH_LOW
  -CATEGORY1_TOKEN, -CATEGORY2_TOKEN,  // 4 = CAT_ONE
  12, 14,                              // 5 = CAT_THREEFOUR
  -CATEGORY3_TOKEN, -CATEGORY4_TOKEN,  // 6 = CAT_THREE
  -CATEGORY5_TOKEN, -CATEGORY6_TOKEN   // 7 = CAT_FIVE
};
/* clang-format on */

const aom_prob av1_cat1_prob[] = { 159 };
const aom_prob av1_cat2_prob[] = { 165, 145 };
const aom_prob av1_cat3_prob[] = { 173, 148, 140 };
const aom_prob av1_cat4_prob[] = { 176, 155, 140, 135 };
const aom_prob av1_cat5_prob[] = { 180, 157, 141, 134, 130 };
const aom_prob av1_cat6_prob[] = { 254, 254, 254, 252, 249, 243, 230,
                                   196, 177, 153, 140, 133, 130, 129 };
#if CONFIG_AOM_HIGHBITDEPTH
const aom_prob av1_cat1_prob_high10[] = { 159 };
const aom_prob av1_cat2_prob_high10[] = { 165, 145 };
const aom_prob av1_cat3_prob_high10[] = { 173, 148, 140 };
const aom_prob av1_cat4_prob_high10[] = { 176, 155, 140, 135 };
const aom_prob av1_cat5_prob_high10[] = { 180, 157, 141, 134, 130 };
const aom_prob av1_cat6_prob_high10[] = {
  255, 255, 254, 254, 254, 252, 249, 243, 230, 196, 177, 153, 140, 133, 130, 129
};
const aom_prob av1_cat1_prob_high12[] = { 159 };
const aom_prob av1_cat2_prob_high12[] = { 165, 145 };
const aom_prob av1_cat3_prob_high12[] = { 173, 148, 140 };
const aom_prob av1_cat4_prob_high12[] = { 176, 155, 140, 135 };
const aom_prob av1_cat5_prob_high12[] = { 180, 157, 141, 134, 130 };
const aom_prob av1_cat6_prob_high12[] = { 255, 255, 255, 255, 254, 254,
                                          254, 252, 249, 243, 230, 196,
                                          177, 153, 140, 133, 130, 129 };
#endif

const uint16_t band_count_table[TX_SIZES_ALL][8] = {
  { 1, 2, 3, 4, 3, 16 - 13, 0 },   { 1, 2, 3, 4, 11, 64 - 21, 0 },
  { 1, 2, 3, 4, 11, 256 - 21, 0 }, { 1, 2, 3, 4, 11, 1024 - 21, 0 },
#if CONFIG_EXT_TX
  { 1, 2, 3, 4, 8, 32 - 18, 0 },   { 1, 2, 3, 4, 8, 32 - 18, 0 },
  { 1, 2, 3, 4, 11, 128 - 21, 0 }, { 1, 2, 3, 4, 11, 128 - 21, 0 },
  { 1, 2, 3, 4, 11, 512 - 21, 0 }, { 1, 2, 3, 4, 11, 512 - 21, 0 },
#endif  // CONFIG_EXT_TX
};

const uint16_t band_cum_count_table[TX_SIZES_ALL][8] = {
  { 0, 1, 3, 6, 10, 13, 16, 0 },  { 0, 1, 3, 6, 10, 21, 64, 0 },
  { 0, 1, 3, 6, 10, 21, 256, 0 }, { 0, 1, 3, 6, 10, 21, 1024, 0 },
#if CONFIG_EXT_TX
  { 0, 1, 3, 6, 10, 18, 32, 0 },  { 0, 1, 3, 6, 10, 18, 32, 0 },
  { 0, 1, 3, 6, 10, 21, 128, 0 }, { 0, 1, 3, 6, 10, 21, 128, 0 },
  { 0, 1, 3, 6, 10, 21, 512, 0 }, { 0, 1, 3, 6, 10, 21, 512, 0 },
#endif  // CONFIG_EXT_TX
};

const uint8_t av1_coefband_trans_8x8plus[1024] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5,
  // beyond MAXBAND_INDEX+1 all values are filled as 5
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
};

#if CONFIG_EXT_TX
const uint8_t av1_coefband_trans_4x8_8x4[32] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4,
  4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
};
#endif  // CONFIG_EXT_TX

const uint8_t av1_coefband_trans_4x4[16] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 5, 5, 5,
};

const uint8_t av1_pt_energy_class[ENTROPY_TOKENS] = { 0, 1, 2, 3, 3, 4,
                                                      4, 5, 5, 5, 5, 5 };

// Model obtained from a 2-sided zero-centered distribution derived
// from a Pareto distribution. The cdf of the distribution is:
// cdf(x) = 0.5 + 0.5 * sgn(x) * [1 - {alpha/(alpha + |x|)} ^ beta]
//
// For a given beta and a given probablity of the 1-node, the alpha
// is first solved, and then the {alpha, beta} pair is used to generate
// the probabilities for the rest of the nodes.

// beta = 8

// Every odd line in this table can be generated from the even lines
// by averaging :
// av1_pareto8_full[l][node] = (av1_pareto8_full[l-1][node] +
//                              av1_pareto8_full[l+1][node] ) >> 1;
const aom_prob av1_pareto8_full[COEFF_PROB_MODELS][MODEL_NODES] = {
  { 3, 86, 128, 6, 86, 23, 88, 29 },
  { 6, 86, 128, 11, 87, 42, 91, 52 },
  { 9, 86, 129, 17, 88, 61, 94, 76 },
  { 12, 86, 129, 22, 88, 77, 97, 93 },
  { 15, 87, 129, 28, 89, 93, 100, 110 },
  { 17, 87, 129, 33, 90, 105, 103, 123 },
  { 20, 88, 130, 38, 91, 118, 106, 136 },
  { 23, 88, 130, 43, 91, 128, 108, 146 },
  { 26, 89, 131, 48, 92, 139, 111, 156 },
  { 28, 89, 131, 53, 93, 147, 114, 163 },
  { 31, 90, 131, 58, 94, 156, 117, 171 },
  { 34, 90, 131, 62, 94, 163, 119, 177 },
  { 37, 90, 132, 66, 95, 171, 122, 184 },
  { 39, 90, 132, 70, 96, 177, 124, 189 },
  { 42, 91, 132, 75, 97, 183, 127, 194 },
  { 44, 91, 132, 79, 97, 188, 129, 198 },
  { 47, 92, 133, 83, 98, 193, 132, 202 },
  { 49, 92, 133, 86, 99, 197, 134, 205 },
  { 52, 93, 133, 90, 100, 201, 137, 208 },
  { 54, 93, 133, 94, 100, 204, 139, 211 },
  { 57, 94, 134, 98, 101, 208, 142, 214 },
  { 59, 94, 134, 101, 102, 211, 144, 216 },
  { 62, 94, 135, 105, 103, 214, 146, 218 },
  { 64, 94, 135, 108, 103, 216, 148, 220 },
  { 66, 95, 135, 111, 104, 219, 151, 222 },
  { 68, 95, 135, 114, 105, 221, 153, 223 },
  { 71, 96, 136, 117, 106, 224, 155, 225 },
  { 73, 96, 136, 120, 106, 225, 157, 226 },
  { 76, 97, 136, 123, 107, 227, 159, 228 },
  { 78, 97, 136, 126, 108, 229, 160, 229 },
  { 80, 98, 137, 129, 109, 231, 162, 231 },
  { 82, 98, 137, 131, 109, 232, 164, 232 },
  { 84, 98, 138, 134, 110, 234, 166, 233 },
  { 86, 98, 138, 137, 111, 235, 168, 234 },
  { 89, 99, 138, 140, 112, 236, 170, 235 },
  { 91, 99, 138, 142, 112, 237, 171, 235 },
  { 93, 100, 139, 145, 113, 238, 173, 236 },
  { 95, 100, 139, 147, 114, 239, 174, 237 },
  { 97, 101, 140, 149, 115, 240, 176, 238 },
  { 99, 101, 140, 151, 115, 241, 177, 238 },
  { 101, 102, 140, 154, 116, 242, 179, 239 },
  { 103, 102, 140, 156, 117, 242, 180, 239 },
  { 105, 103, 141, 158, 118, 243, 182, 240 },
  { 107, 103, 141, 160, 118, 243, 183, 240 },
  { 109, 104, 141, 162, 119, 244, 185, 241 },
  { 111, 104, 141, 164, 119, 244, 186, 241 },
  { 113, 104, 142, 166, 120, 245, 187, 242 },
  { 114, 104, 142, 168, 121, 245, 188, 242 },
  { 116, 105, 143, 170, 122, 246, 190, 243 },
  { 118, 105, 143, 171, 122, 246, 191, 243 },
  { 120, 106, 143, 173, 123, 247, 192, 244 },
  { 121, 106, 143, 175, 124, 247, 193, 244 },
  { 123, 107, 144, 177, 125, 248, 195, 244 },
  { 125, 107, 144, 178, 125, 248, 196, 244 },
  { 127, 108, 145, 180, 126, 249, 197, 245 },
  { 128, 108, 145, 181, 127, 249, 198, 245 },
  { 130, 109, 145, 183, 128, 249, 199, 245 },
  { 132, 109, 145, 184, 128, 249, 200, 245 },
  { 134, 110, 146, 186, 129, 250, 201, 246 },
  { 135, 110, 146, 187, 130, 250, 202, 246 },
  { 137, 111, 147, 189, 131, 251, 203, 246 },
  { 138, 111, 147, 190, 131, 251, 204, 246 },
  { 140, 112, 147, 192, 132, 251, 205, 247 },
  { 141, 112, 147, 193, 132, 251, 206, 247 },
  { 143, 113, 148, 194, 133, 251, 207, 247 },
  { 144, 113, 148, 195, 134, 251, 207, 247 },
  { 146, 114, 149, 197, 135, 252, 208, 248 },
  { 147, 114, 149, 198, 135, 252, 209, 248 },
  { 149, 115, 149, 199, 136, 252, 210, 248 },
  { 150, 115, 149, 200, 137, 252, 210, 248 },
  { 152, 115, 150, 201, 138, 252, 211, 248 },
  { 153, 115, 150, 202, 138, 252, 212, 248 },
  { 155, 116, 151, 204, 139, 253, 213, 249 },
  { 156, 116, 151, 205, 139, 253, 213, 249 },
  { 158, 117, 151, 206, 140, 253, 214, 249 },
  { 159, 117, 151, 207, 141, 253, 215, 249 },
  { 161, 118, 152, 208, 142, 253, 216, 249 },
  { 162, 118, 152, 209, 142, 253, 216, 249 },
  { 163, 119, 153, 210, 143, 253, 217, 249 },
  { 164, 119, 153, 211, 143, 253, 217, 249 },
  { 166, 120, 153, 212, 144, 254, 218, 250 },
  { 167, 120, 153, 212, 145, 254, 219, 250 },
  { 168, 121, 154, 213, 146, 254, 220, 250 },
  { 169, 121, 154, 214, 146, 254, 220, 250 },
  { 171, 122, 155, 215, 147, 254, 221, 250 },
  { 172, 122, 155, 216, 147, 254, 221, 250 },
  { 173, 123, 155, 217, 148, 254, 222, 250 },
  { 174, 123, 155, 217, 149, 254, 222, 250 },
  { 176, 124, 156, 218, 150, 254, 223, 250 },
  { 177, 124, 156, 219, 150, 254, 223, 250 },
  { 178, 125, 157, 220, 151, 254, 224, 251 },
  { 179, 125, 157, 220, 151, 254, 224, 251 },
  { 180, 126, 157, 221, 152, 254, 225, 251 },
  { 181, 126, 157, 221, 152, 254, 225, 251 },
  { 183, 127, 158, 222, 153, 254, 226, 251 },
  { 184, 127, 158, 223, 154, 254, 226, 251 },
  { 185, 128, 159, 224, 155, 255, 227, 251 },
  { 186, 128, 159, 224, 155, 255, 227, 251 },
  { 187, 129, 160, 225, 156, 255, 228, 251 },
  { 188, 130, 160, 225, 156, 255, 228, 251 },
  { 189, 131, 160, 226, 157, 255, 228, 251 },
  { 190, 131, 160, 226, 158, 255, 228, 251 },
  { 191, 132, 161, 227, 159, 255, 229, 251 },
  { 192, 132, 161, 227, 159, 255, 229, 251 },
  { 193, 133, 162, 228, 160, 255, 230, 252 },
  { 194, 133, 162, 229, 160, 255, 230, 252 },
  { 195, 134, 163, 230, 161, 255, 231, 252 },
  { 196, 134, 163, 230, 161, 255, 231, 252 },
  { 197, 135, 163, 231, 162, 255, 231, 252 },
  { 198, 135, 163, 231, 162, 255, 231, 252 },
  { 199, 136, 164, 232, 163, 255, 232, 252 },
  { 200, 136, 164, 232, 164, 255, 232, 252 },
  { 201, 137, 165, 233, 165, 255, 233, 252 },
  { 201, 137, 165, 233, 165, 255, 233, 252 },
  { 202, 138, 166, 233, 166, 255, 233, 252 },
  { 203, 138, 166, 233, 166, 255, 233, 252 },
  { 204, 139, 166, 234, 167, 255, 234, 252 },
  { 205, 139, 166, 234, 167, 255, 234, 252 },
  { 206, 140, 167, 235, 168, 255, 235, 252 },
  { 206, 140, 167, 235, 168, 255, 235, 252 },
  { 207, 141, 168, 236, 169, 255, 235, 252 },
  { 208, 141, 168, 236, 170, 255, 235, 252 },
  { 209, 142, 169, 237, 171, 255, 236, 252 },
  { 209, 143, 169, 237, 171, 255, 236, 252 },
  { 210, 144, 169, 237, 172, 255, 236, 252 },
  { 211, 144, 169, 237, 172, 255, 236, 252 },
  { 212, 145, 170, 238, 173, 255, 237, 252 },
  { 213, 145, 170, 238, 173, 255, 237, 252 },
  { 214, 146, 171, 239, 174, 255, 237, 253 },
  { 214, 146, 171, 239, 174, 255, 237, 253 },
  { 215, 147, 172, 240, 175, 255, 238, 253 },
  { 215, 147, 172, 240, 175, 255, 238, 253 },
  { 216, 148, 173, 240, 176, 255, 238, 253 },
  { 217, 148, 173, 240, 176, 255, 238, 253 },
  { 218, 149, 173, 241, 177, 255, 239, 253 },
  { 218, 149, 173, 241, 178, 255, 239, 253 },
  { 219, 150, 174, 241, 179, 255, 239, 253 },
  { 219, 151, 174, 241, 179, 255, 239, 253 },
  { 220, 152, 175, 242, 180, 255, 240, 253 },
  { 221, 152, 175, 242, 180, 255, 240, 253 },
  { 222, 153, 176, 242, 181, 255, 240, 253 },
  { 222, 153, 176, 242, 181, 255, 240, 253 },
  { 223, 154, 177, 243, 182, 255, 240, 253 },
  { 223, 154, 177, 243, 182, 255, 240, 253 },
  { 224, 155, 178, 244, 183, 255, 241, 253 },
  { 224, 155, 178, 244, 183, 255, 241, 253 },
  { 225, 156, 178, 244, 184, 255, 241, 253 },
  { 225, 157, 178, 244, 184, 255, 241, 253 },
  { 226, 158, 179, 244, 185, 255, 242, 253 },
  { 227, 158, 179, 244, 185, 255, 242, 253 },
  { 228, 159, 180, 245, 186, 255, 242, 253 },
  { 228, 159, 180, 245, 186, 255, 242, 253 },
  { 229, 160, 181, 245, 187, 255, 242, 253 },
  { 229, 160, 181, 245, 187, 255, 242, 253 },
  { 230, 161, 182, 246, 188, 255, 243, 253 },
  { 230, 162, 182, 246, 188, 255, 243, 253 },
  { 231, 163, 183, 246, 189, 255, 243, 253 },
  { 231, 163, 183, 246, 189, 255, 243, 253 },
  { 232, 164, 184, 247, 190, 255, 243, 253 },
  { 232, 164, 184, 247, 190, 255, 243, 253 },
  { 233, 165, 185, 247, 191, 255, 244, 253 },
  { 233, 165, 185, 247, 191, 255, 244, 253 },
  { 234, 166, 185, 247, 192, 255, 244, 253 },
  { 234, 167, 185, 247, 192, 255, 244, 253 },
  { 235, 168, 186, 248, 193, 255, 244, 253 },
  { 235, 168, 186, 248, 193, 255, 244, 253 },
  { 236, 169, 187, 248, 194, 255, 244, 253 },
  { 236, 169, 187, 248, 194, 255, 244, 253 },
  { 236, 170, 188, 248, 195, 255, 245, 253 },
  { 236, 170, 188, 248, 195, 255, 245, 253 },
  { 237, 171, 189, 249, 196, 255, 245, 254 },
  { 237, 172, 189, 249, 196, 255, 245, 254 },
  { 238, 173, 190, 249, 197, 255, 245, 254 },
  { 238, 173, 190, 249, 197, 255, 245, 254 },
  { 239, 174, 191, 249, 198, 255, 245, 254 },
  { 239, 174, 191, 249, 198, 255, 245, 254 },
  { 240, 175, 192, 249, 199, 255, 246, 254 },
  { 240, 176, 192, 249, 199, 255, 246, 254 },
  { 240, 177, 193, 250, 200, 255, 246, 254 },
  { 240, 177, 193, 250, 200, 255, 246, 254 },
  { 241, 178, 194, 250, 201, 255, 246, 254 },
  { 241, 178, 194, 250, 201, 255, 246, 254 },
  { 242, 179, 195, 250, 202, 255, 246, 254 },
  { 242, 180, 195, 250, 202, 255, 246, 254 },
  { 242, 181, 196, 250, 203, 255, 247, 254 },
  { 242, 181, 196, 250, 203, 255, 247, 254 },
  { 243, 182, 197, 251, 204, 255, 247, 254 },
  { 243, 183, 197, 251, 204, 255, 247, 254 },
  { 244, 184, 198, 251, 205, 255, 247, 254 },
  { 244, 184, 198, 251, 205, 255, 247, 254 },
  { 244, 185, 199, 251, 206, 255, 247, 254 },
  { 244, 185, 199, 251, 206, 255, 247, 254 },
  { 245, 186, 200, 251, 207, 255, 247, 254 },
  { 245, 187, 200, 251, 207, 255, 247, 254 },
  { 246, 188, 201, 252, 207, 255, 248, 254 },
  { 246, 188, 201, 252, 207, 255, 248, 254 },
  { 246, 189, 202, 252, 208, 255, 248, 254 },
  { 246, 190, 202, 252, 208, 255, 248, 254 },
  { 247, 191, 203, 252, 209, 255, 248, 254 },
  { 247, 191, 203, 252, 209, 255, 248, 254 },
  { 247, 192, 204, 252, 210, 255, 248, 254 },
  { 247, 193, 204, 252, 210, 255, 248, 254 },
  { 248, 194, 205, 252, 211, 255, 248, 254 },
  { 248, 194, 205, 252, 211, 255, 248, 254 },
  { 248, 195, 206, 252, 212, 255, 249, 254 },
  { 248, 196, 206, 252, 212, 255, 249, 254 },
  { 249, 197, 207, 253, 213, 255, 249, 254 },
  { 249, 197, 207, 253, 213, 255, 249, 254 },
  { 249, 198, 208, 253, 214, 255, 249, 254 },
  { 249, 199, 209, 253, 214, 255, 249, 254 },
  { 250, 200, 210, 253, 215, 255, 249, 254 },
  { 250, 200, 210, 253, 215, 255, 249, 254 },
  { 250, 201, 211, 253, 215, 255, 249, 254 },
  { 250, 202, 211, 253, 215, 255, 249, 254 },
  { 250, 203, 212, 253, 216, 255, 249, 254 },
  { 250, 203, 212, 253, 216, 255, 249, 254 },
  { 251, 204, 213, 253, 217, 255, 250, 254 },
  { 251, 205, 213, 253, 217, 255, 250, 254 },
  { 251, 206, 214, 254, 218, 255, 250, 254 },
  { 251, 206, 215, 254, 218, 255, 250, 254 },
  { 252, 207, 216, 254, 219, 255, 250, 254 },
  { 252, 208, 216, 254, 219, 255, 250, 254 },
  { 252, 209, 217, 254, 220, 255, 250, 254 },
  { 252, 210, 217, 254, 220, 255, 250, 254 },
  { 252, 211, 218, 254, 221, 255, 250, 254 },
  { 252, 212, 218, 254, 221, 255, 250, 254 },
  { 253, 213, 219, 254, 222, 255, 250, 254 },
  { 253, 213, 220, 254, 222, 255, 250, 254 },
  { 253, 214, 221, 254, 223, 255, 250, 254 },
  { 253, 215, 221, 254, 223, 255, 250, 254 },
  { 253, 216, 222, 254, 224, 255, 251, 254 },
  { 253, 217, 223, 254, 224, 255, 251, 254 },
  { 253, 218, 224, 254, 225, 255, 251, 254 },
  { 253, 219, 224, 254, 225, 255, 251, 254 },
  { 254, 220, 225, 254, 225, 255, 251, 254 },
  { 254, 221, 226, 254, 225, 255, 251, 254 },
  { 254, 222, 227, 255, 226, 255, 251, 254 },
  { 254, 223, 227, 255, 226, 255, 251, 254 },
  { 254, 224, 228, 255, 227, 255, 251, 254 },
  { 254, 225, 229, 255, 227, 255, 251, 254 },
  { 254, 226, 230, 255, 228, 255, 251, 254 },
  { 254, 227, 230, 255, 229, 255, 251, 254 },
  { 255, 228, 231, 255, 230, 255, 251, 254 },
  { 255, 229, 232, 255, 230, 255, 251, 254 },
  { 255, 230, 233, 255, 231, 255, 252, 254 },
  { 255, 231, 234, 255, 231, 255, 252, 254 },
  { 255, 232, 235, 255, 232, 255, 252, 254 },
  { 255, 233, 236, 255, 232, 255, 252, 254 },
  { 255, 235, 237, 255, 233, 255, 252, 254 },
  { 255, 236, 238, 255, 234, 255, 252, 254 },
  { 255, 238, 240, 255, 235, 255, 252, 255 },
  { 255, 239, 241, 255, 235, 255, 252, 254 },
  { 255, 241, 243, 255, 236, 255, 252, 254 },
  { 255, 243, 245, 255, 237, 255, 252, 254 },
  { 255, 246, 247, 255, 239, 255, 253, 255 },
};

#if CONFIG_ANS
// Model obtained from a 2-sided zero-centerd distribuition derived
// from a Pareto distribution. The cdf of the distribution is:
// cdf(x) = 0.5 + 0.5 * sgn(x) * [1 - {alpha/(alpha + |x|)} ^ beta]
//
// For a given beta and a given probablity of the 1-node, the alpha
// is first solved, and then the {alpha, beta} pair is used to generate
// the probabilities for the rest of the nodes.
//
// beta = 8
// Values for tokens ONE_TOKEN through CATEGORY6_TOKEN included here.
// ZERO_TOKEN and EOB_TOKEN are coded as flags outside this coder.
const AnsP10 av1_pareto8_token_probs[COEFF_PROB_MODELS][ENTROPY_TOKENS - 2] = {
  { 4, 4, 4, 4, 8, 15, 30, 57, 103, 795 },
  { 8, 8, 8, 8, 15, 30, 57, 103, 168, 619 },
  { 12, 12, 12, 12, 23, 43, 80, 138, 205, 487 },
  { 16, 16, 15, 15, 30, 56, 101, 165, 225, 385 },
  { 20, 20, 19, 19, 36, 68, 119, 186, 231, 306 },
  { 24, 23, 23, 22, 43, 79, 135, 201, 230, 244 },
  { 28, 27, 26, 26, 49, 89, 149, 211, 223, 196 },
  { 32, 31, 30, 29, 55, 98, 160, 218, 212, 159 },
  { 36, 35, 33, 32, 60, 107, 171, 221, 200, 129 },
  { 40, 38, 37, 35, 66, 115, 179, 222, 187, 105 },
  { 44, 42, 40, 38, 71, 122, 186, 221, 174, 86 },
  { 48, 45, 43, 41, 76, 129, 192, 219, 160, 71 },
  { 52, 49, 46, 44, 80, 136, 196, 215, 148, 58 },
  { 56, 53, 49, 46, 85, 142, 200, 210, 135, 48 },
  { 60, 56, 52, 49, 89, 147, 203, 204, 124, 40 },
  { 64, 60, 55, 52, 93, 151, 205, 198, 113, 33 },
  { 68, 63, 58, 54, 97, 156, 205, 192, 103, 28 },
  { 72, 66, 61, 57, 100, 160, 206, 185, 94, 23 },
  { 76, 70, 64, 59, 104, 163, 205, 178, 85, 20 },
  { 80, 73, 67, 61, 107, 166, 205, 171, 77, 17 },
  { 84, 76, 69, 63, 110, 169, 204, 164, 71, 14 },
  { 88, 80, 72, 65, 113, 171, 202, 157, 64, 12 },
  { 92, 83, 75, 67, 116, 173, 200, 150, 58, 10 },
  { 96, 86, 77, 69, 118, 175, 198, 143, 53, 9 },
  { 100, 89, 80, 71, 121, 176, 195, 137, 48, 7 },
  { 104, 92, 82, 73, 123, 178, 192, 130, 44, 6 },
  { 108, 96, 84, 75, 125, 178, 189, 124, 40, 5 },
  { 112, 98, 87, 76, 127, 179, 186, 118, 36, 5 },
  { 116, 101, 89, 78, 129, 179, 183, 112, 33, 4 },
  { 120, 104, 91, 80, 131, 180, 179, 106, 30, 3 },
  { 124, 107, 93, 81, 132, 180, 176, 101, 27, 3 },
  { 128, 110, 95, 82, 134, 179, 172, 96, 25, 3 },
  { 132, 113, 97, 84, 135, 179, 168, 91, 23, 2 },
  { 136, 116, 99, 85, 136, 179, 164, 86, 21, 2 },
  { 140, 119, 101, 86, 137, 178, 160, 82, 19, 2 },
  { 144, 122, 103, 88, 138, 177, 157, 77, 17, 1 },
  { 148, 124, 105, 89, 139, 176, 153, 73, 16, 1 },
  { 152, 127, 107, 90, 140, 175, 149, 69, 14, 1 },
  { 156, 130, 108, 91, 141, 173, 145, 66, 13, 1 },
  { 160, 133, 110, 92, 141, 172, 141, 62, 12, 1 },
  { 164, 135, 111, 93, 142, 171, 137, 59, 11, 1 },
  { 168, 138, 113, 94, 142, 169, 133, 56, 10, 1 },
  { 172, 140, 115, 94, 142, 168, 130, 53, 9, 1 },
  { 176, 143, 116, 95, 143, 166, 126, 50, 8, 1 },
  { 180, 145, 118, 96, 143, 164, 122, 47, 8, 1 },
  { 184, 147, 119, 96, 143, 163, 119, 45, 7, 1 },
  { 188, 150, 120, 97, 143, 161, 116, 42, 6, 1 },
  { 192, 152, 121, 98, 143, 159, 112, 40, 6, 1 },
  { 196, 155, 123, 98, 142, 157, 109, 38, 5, 1 },
  { 200, 157, 124, 99, 142, 155, 105, 36, 5, 1 },
  { 204, 159, 125, 99, 142, 153, 102, 34, 5, 1 },
  { 208, 161, 126, 100, 142, 151, 99, 32, 4, 1 },
  { 212, 164, 127, 100, 141, 149, 96, 30, 4, 1 },
  { 216, 166, 129, 100, 141, 147, 93, 28, 3, 1 },
  { 220, 168, 130, 101, 140, 144, 90, 27, 3, 1 },
  { 224, 170, 131, 101, 140, 142, 87, 25, 3, 1 },
  { 228, 172, 132, 101, 139, 140, 84, 24, 3, 1 },
  { 232, 174, 132, 101, 139, 138, 81, 23, 3, 1 },
  { 236, 176, 133, 101, 138, 136, 79, 22, 2, 1 },
  { 240, 178, 134, 102, 137, 134, 76, 20, 2, 1 },
  { 244, 180, 135, 102, 136, 131, 74, 19, 2, 1 },
  { 248, 182, 135, 102, 136, 129, 71, 18, 2, 1 },
  { 252, 184, 136, 101, 135, 127, 69, 17, 2, 1 },
  { 256, 186, 137, 102, 134, 124, 66, 16, 2, 1 },
  { 260, 188, 138, 102, 133, 122, 64, 15, 1, 1 },
  { 264, 190, 138, 101, 132, 120, 62, 15, 1, 1 },
  { 268, 191, 139, 101, 131, 118, 60, 14, 1, 1 },
  { 272, 193, 139, 101, 130, 116, 58, 13, 1, 1 },
  { 276, 195, 139, 101, 129, 114, 56, 12, 1, 1 },
  { 280, 196, 140, 101, 128, 111, 54, 12, 1, 1 },
  { 284, 198, 140, 101, 127, 109, 52, 11, 1, 1 },
  { 288, 200, 141, 100, 126, 107, 50, 10, 1, 1 },
  { 292, 201, 141, 100, 125, 105, 48, 10, 1, 1 },
  { 296, 203, 141, 100, 123, 103, 47, 9, 1, 1 },
  { 300, 204, 142, 99, 122, 101, 45, 9, 1, 1 },
  { 304, 206, 142, 99, 121, 99, 43, 8, 1, 1 },
  { 308, 207, 142, 99, 119, 97, 42, 8, 1, 1 },
  { 312, 209, 142, 99, 118, 95, 40, 7, 1, 1 },
  { 316, 210, 142, 98, 117, 93, 39, 7, 1, 1 },
  { 320, 211, 142, 98, 116, 91, 37, 7, 1, 1 },
  { 324, 213, 142, 97, 115, 89, 36, 6, 1, 1 },
  { 328, 214, 142, 97, 113, 87, 35, 6, 1, 1 },
  { 332, 215, 143, 96, 112, 85, 33, 6, 1, 1 },
  { 336, 216, 143, 96, 111, 83, 32, 5, 1, 1 },
  { 340, 218, 143, 95, 109, 81, 31, 5, 1, 1 },
  { 344, 219, 142, 95, 108, 79, 30, 5, 1, 1 },
  { 348, 220, 142, 94, 107, 78, 29, 4, 1, 1 },
  { 352, 221, 142, 94, 105, 76, 28, 4, 1, 1 },
  { 356, 222, 142, 93, 104, 74, 27, 4, 1, 1 },
  { 360, 223, 142, 92, 103, 72, 26, 4, 1, 1 },
  { 364, 224, 142, 92, 101, 70, 25, 4, 1, 1 },
  { 368, 225, 142, 91, 100, 69, 24, 3, 1, 1 },
  { 372, 226, 141, 91, 99, 67, 23, 3, 1, 1 },
  { 376, 227, 141, 90, 97, 66, 22, 3, 1, 1 },
  { 380, 228, 141, 89, 96, 64, 21, 3, 1, 1 },
  { 384, 229, 140, 89, 95, 62, 20, 3, 1, 1 },
  { 388, 229, 140, 88, 93, 61, 20, 3, 1, 1 },
  { 392, 230, 140, 87, 92, 60, 19, 2, 1, 1 },
  { 396, 231, 140, 86, 91, 58, 18, 2, 1, 1 },
  { 400, 232, 139, 86, 89, 57, 17, 2, 1, 1 },
  { 404, 232, 139, 85, 88, 55, 17, 2, 1, 1 },
  { 408, 233, 138, 84, 87, 54, 16, 2, 1, 1 },
  { 412, 234, 138, 84, 85, 52, 15, 2, 1, 1 },
  { 416, 234, 137, 83, 84, 51, 15, 2, 1, 1 },
  { 420, 235, 137, 82, 82, 50, 14, 2, 1, 1 },
  { 424, 236, 136, 81, 81, 48, 14, 2, 1, 1 },
  { 428, 236, 136, 81, 80, 47, 13, 1, 1, 1 },
  { 432, 236, 135, 80, 79, 46, 13, 1, 1, 1 },
  { 436, 237, 135, 79, 77, 45, 12, 1, 1, 1 },
  { 440, 238, 134, 78, 76, 43, 12, 1, 1, 1 },
  { 444, 238, 134, 77, 75, 42, 11, 1, 1, 1 },
  { 448, 238, 133, 77, 73, 41, 11, 1, 1, 1 },
  { 452, 239, 132, 76, 72, 40, 10, 1, 1, 1 },
  { 456, 239, 131, 75, 71, 39, 10, 1, 1, 1 },
  { 460, 239, 131, 74, 70, 38, 9, 1, 1, 1 },
  { 464, 240, 130, 73, 68, 37, 9, 1, 1, 1 },
  { 468, 240, 129, 72, 67, 36, 9, 1, 1, 1 },
  { 472, 240, 128, 72, 66, 35, 8, 1, 1, 1 },
  { 476, 240, 127, 71, 65, 34, 8, 1, 1, 1 },
  { 480, 240, 127, 70, 63, 33, 8, 1, 1, 1 },
  { 484, 241, 126, 69, 62, 32, 7, 1, 1, 1 },
  { 488, 241, 125, 68, 61, 31, 7, 1, 1, 1 },
  { 492, 241, 124, 67, 60, 30, 7, 1, 1, 1 },
  { 496, 241, 124, 66, 59, 29, 6, 1, 1, 1 },
  { 500, 240, 123, 66, 58, 28, 6, 1, 1, 1 },
  { 504, 240, 122, 65, 57, 27, 6, 1, 1, 1 },
  { 508, 240, 121, 64, 55, 27, 6, 1, 1, 1 },
  { 512, 241, 120, 63, 54, 26, 5, 1, 1, 1 },
  { 516, 241, 119, 62, 53, 25, 5, 1, 1, 1 },
  { 520, 240, 118, 62, 52, 24, 5, 1, 1, 1 },
  { 524, 240, 117, 60, 51, 24, 5, 1, 1, 1 },
  { 528, 239, 116, 60, 50, 23, 5, 1, 1, 1 },
  { 532, 239, 116, 59, 49, 22, 4, 1, 1, 1 },
  { 536, 239, 115, 58, 48, 21, 4, 1, 1, 1 },
  { 540, 239, 113, 57, 47, 21, 4, 1, 1, 1 },
  { 544, 238, 113, 56, 46, 20, 4, 1, 1, 1 },
  { 548, 238, 112, 55, 45, 19, 4, 1, 1, 1 },
  { 552, 238, 110, 55, 44, 19, 3, 1, 1, 1 },
  { 556, 237, 110, 54, 43, 18, 3, 1, 1, 1 },
  { 560, 237, 108, 53, 42, 18, 3, 1, 1, 1 },
  { 564, 236, 108, 52, 41, 17, 3, 1, 1, 1 },
  { 568, 236, 106, 51, 40, 17, 3, 1, 1, 1 },
  { 572, 235, 105, 51, 39, 16, 3, 1, 1, 1 },
  { 576, 235, 104, 50, 38, 15, 3, 1, 1, 1 },
  { 580, 234, 103, 49, 37, 15, 3, 1, 1, 1 },
  { 584, 234, 102, 48, 37, 14, 2, 1, 1, 1 },
  { 588, 233, 101, 47, 36, 14, 2, 1, 1, 1 },
  { 592, 233, 100, 46, 35, 13, 2, 1, 1, 1 },
  { 596, 231, 99, 46, 34, 13, 2, 1, 1, 1 },
  { 600, 230, 98, 45, 33, 13, 2, 1, 1, 1 },
  { 604, 230, 97, 44, 32, 12, 2, 1, 1, 1 },
  { 608, 229, 96, 43, 31, 12, 2, 1, 1, 1 },
  { 612, 228, 95, 42, 31, 11, 2, 1, 1, 1 },
  { 616, 227, 93, 42, 30, 11, 2, 1, 1, 1 },
  { 620, 227, 92, 41, 29, 10, 2, 1, 1, 1 },
  { 624, 226, 92, 40, 28, 10, 1, 1, 1, 1 },
  { 628, 225, 90, 39, 28, 10, 1, 1, 1, 1 },
  { 632, 224, 89, 39, 27, 9, 1, 1, 1, 1 },
  { 636, 223, 88, 38, 26, 9, 1, 1, 1, 1 },
  { 640, 222, 87, 37, 25, 9, 1, 1, 1, 1 },
  { 644, 221, 86, 36, 25, 8, 1, 1, 1, 1 },
  { 648, 220, 84, 36, 24, 8, 1, 1, 1, 1 },
  { 652, 219, 83, 35, 23, 8, 1, 1, 1, 1 },
  { 656, 218, 82, 34, 23, 7, 1, 1, 1, 1 },
  { 660, 217, 81, 33, 22, 7, 1, 1, 1, 1 },
  { 664, 215, 80, 33, 21, 7, 1, 1, 1, 1 },
  { 668, 214, 78, 32, 21, 7, 1, 1, 1, 1 },
  { 672, 213, 78, 31, 20, 6, 1, 1, 1, 1 },
  { 676, 211, 76, 31, 20, 6, 1, 1, 1, 1 },
  { 680, 210, 75, 30, 19, 6, 1, 1, 1, 1 },
  { 684, 209, 74, 29, 18, 6, 1, 1, 1, 1 },
  { 688, 208, 73, 28, 18, 5, 1, 1, 1, 1 },
  { 692, 206, 72, 28, 17, 5, 1, 1, 1, 1 },
  { 696, 205, 70, 27, 17, 5, 1, 1, 1, 1 },
  { 700, 203, 69, 27, 16, 5, 1, 1, 1, 1 },
  { 704, 201, 68, 26, 16, 5, 1, 1, 1, 1 },
  { 708, 201, 67, 25, 15, 4, 1, 1, 1, 1 },
  { 712, 198, 66, 25, 15, 4, 1, 1, 1, 1 },
  { 716, 197, 65, 24, 14, 4, 1, 1, 1, 1 },
  { 720, 196, 63, 23, 14, 4, 1, 1, 1, 1 },
  { 724, 194, 62, 23, 13, 4, 1, 1, 1, 1 },
  { 728, 193, 61, 22, 13, 3, 1, 1, 1, 1 },
  { 732, 191, 60, 22, 12, 3, 1, 1, 1, 1 },
  { 736, 189, 59, 21, 12, 3, 1, 1, 1, 1 },
  { 740, 188, 58, 20, 11, 3, 1, 1, 1, 1 },
  { 744, 186, 56, 20, 11, 3, 1, 1, 1, 1 },
  { 748, 184, 55, 19, 11, 3, 1, 1, 1, 1 },
  { 752, 182, 54, 19, 10, 3, 1, 1, 1, 1 },
  { 756, 181, 53, 18, 10, 2, 1, 1, 1, 1 },
  { 760, 179, 52, 18, 9, 2, 1, 1, 1, 1 },
  { 764, 177, 51, 17, 9, 2, 1, 1, 1, 1 },
  { 768, 174, 50, 17, 9, 2, 1, 1, 1, 1 },
  { 772, 173, 49, 16, 8, 2, 1, 1, 1, 1 },
  { 776, 171, 47, 16, 8, 2, 1, 1, 1, 1 },
  { 780, 169, 46, 15, 8, 2, 1, 1, 1, 1 },
  { 784, 167, 45, 15, 7, 2, 1, 1, 1, 1 },
  { 788, 165, 44, 14, 7, 2, 1, 1, 1, 1 },
  { 792, 162, 43, 14, 7, 2, 1, 1, 1, 1 },
  { 796, 161, 42, 13, 7, 1, 1, 1, 1, 1 },
  { 800, 159, 41, 13, 6, 1, 1, 1, 1, 1 },
  { 804, 157, 40, 12, 6, 1, 1, 1, 1, 1 },
  { 808, 154, 39, 12, 6, 1, 1, 1, 1, 1 },
  { 812, 153, 38, 11, 5, 1, 1, 1, 1, 1 },
  { 816, 150, 37, 11, 5, 1, 1, 1, 1, 1 },
  { 820, 148, 36, 10, 5, 1, 1, 1, 1, 1 },
  { 824, 145, 35, 10, 5, 1, 1, 1, 1, 1 },
  { 828, 143, 34, 10, 4, 1, 1, 1, 1, 1 },
  { 832, 141, 33, 9, 4, 1, 1, 1, 1, 1 },
  { 836, 138, 32, 9, 4, 1, 1, 1, 1, 1 },
  { 840, 136, 30, 9, 4, 1, 1, 1, 1, 1 },
  { 844, 133, 30, 8, 4, 1, 1, 1, 1, 1 },
  { 848, 131, 29, 8, 3, 1, 1, 1, 1, 1 },
  { 852, 129, 28, 7, 3, 1, 1, 1, 1, 1 },
  { 856, 126, 27, 7, 3, 1, 1, 1, 1, 1 },
  { 860, 123, 26, 7, 3, 1, 1, 1, 1, 1 },
  { 864, 121, 25, 6, 3, 1, 1, 1, 1, 1 },
  { 868, 118, 24, 6, 3, 1, 1, 1, 1, 1 },
  { 872, 116, 23, 6, 2, 1, 1, 1, 1, 1 },
  { 876, 113, 22, 6, 2, 1, 1, 1, 1, 1 },
  { 880, 111, 21, 5, 2, 1, 1, 1, 1, 1 },
  { 884, 108, 20, 5, 2, 1, 1, 1, 1, 1 },
  { 888, 105, 19, 5, 2, 1, 1, 1, 1, 1 },
  { 892, 102, 19, 4, 2, 1, 1, 1, 1, 1 },
  { 896, 99, 18, 4, 2, 1, 1, 1, 1, 1 },
  { 900, 97, 17, 4, 1, 1, 1, 1, 1, 1 },
  { 904, 94, 16, 4, 1, 1, 1, 1, 1, 1 },
  { 908, 92, 15, 3, 1, 1, 1, 1, 1, 1 },
  { 912, 89, 14, 3, 1, 1, 1, 1, 1, 1 },
  { 916, 85, 14, 3, 1, 1, 1, 1, 1, 1 },
  { 920, 82, 13, 3, 1, 1, 1, 1, 1, 1 },
  { 924, 79, 12, 3, 1, 1, 1, 1, 1, 1 },
  { 928, 77, 11, 2, 1, 1, 1, 1, 1, 1 },
  { 932, 73, 11, 2, 1, 1, 1, 1, 1, 1 },
  { 936, 70, 10, 2, 1, 1, 1, 1, 1, 1 },
  { 940, 67, 9, 2, 1, 1, 1, 1, 1, 1 },
  { 944, 64, 8, 2, 1, 1, 1, 1, 1, 1 },
  { 948, 60, 8, 2, 1, 1, 1, 1, 1, 1 },
  { 952, 58, 7, 1, 1, 1, 1, 1, 1, 1 },
  { 956, 54, 7, 1, 1, 1, 1, 1, 1, 1 },
  { 960, 51, 6, 1, 1, 1, 1, 1, 1, 1 },
  { 964, 48, 5, 1, 1, 1, 1, 1, 1, 1 },
  { 968, 44, 5, 1, 1, 1, 1, 1, 1, 1 },
  { 972, 41, 4, 1, 1, 1, 1, 1, 1, 1 },
  { 976, 37, 4, 1, 1, 1, 1, 1, 1, 1 },
  { 980, 34, 3, 1, 1, 1, 1, 1, 1, 1 },
  { 984, 30, 3, 1, 1, 1, 1, 1, 1, 1 },
  { 988, 27, 2, 1, 1, 1, 1, 1, 1, 1 },
  { 992, 23, 2, 1, 1, 1, 1, 1, 1, 1 },
  { 996, 19, 2, 1, 1, 1, 1, 1, 1, 1 },
  { 1000, 16, 1, 1, 1, 1, 1, 1, 1, 1 },
  { 1004, 12, 1, 1, 1, 1, 1, 1, 1, 1 },
  { 1008, 8, 1, 1, 1, 1, 1, 1, 1, 1 },
  { 1012, 4, 1, 1, 1, 1, 1, 1, 1, 1 },
  { 1015, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
  { 1015, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
};
#endif  // CONFIG_ANS

/* clang-format off */
#if CONFIG_ENTROPY
const av1_coeff_probs_model
default_qctx_coef_probs[QCTX_BINS][TX_SIZES][PLANE_TYPES] = {
    {  // Q_Index 0
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {182,  34, 137}, { 79,  39, 103}, { 10,  28,  51},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 45,  88, 147}, { 46,  80, 140}, { 25,  69, 119},
                        { 12,  57,  96}, {  4,  41,  65}, {  1,  20,  31},
                    },
                    {  // band 2
                        { 58, 124, 190}, { 39, 106, 178}, { 16,  86, 147},
                        {  7,  69, 114}, {  3,  50,  80}, {  1,  25,  42},
                    },
                    {  // band 3
                        { 90, 138, 215}, { 54, 116, 198}, { 18,  86, 155},
                        {  5,  62, 112}, {  1,  38,  68}, {  1,  17,  30},
                    },
                    {  // band 4
                        {126, 149, 231}, { 82, 114, 211}, { 21,  80, 157},
                        {  6,  56, 105}, {  1,  36,  64}, {  1,  17,  31},
                    },
                    {  // band 5
                        {171,  56, 236}, {140,  54, 219}, { 57,  45, 167},
                        { 26,  36, 113}, { 11,  29,  72}, {  3,  18,  39},
                    },
                },
                {  // Intra
                    {  // band 0
                        {153, 122, 186}, {106, 109, 171}, { 36,  84, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 27, 151, 201}, { 34, 131, 199}, { 23, 102, 161},
                        { 10,  80, 120}, {  4,  52,  78}, {  1,  24,  37},
                    },
                    {  // band 2
                        { 43, 158, 213}, { 35, 133, 203}, {  8,  92, 151},
                        {  2,  64, 106}, {  1,  36,  60}, {  1,  13,  24},
                    },
                    {  // band 3
                        { 68, 167, 223}, { 36, 135, 211}, {  9,  94, 157},
                        {  2,  67, 112}, {  1,  40,  68}, {  1,  17,  31},
                    },
                    {  // band 4
                        {131, 146, 237}, { 72, 119, 223}, { 17,  82, 164},
                        {  4,  55, 107}, {  1,  34,  63}, {  1,  16,  29},
                    },
                    {  // band 5
                        {184,  68, 244}, {153,  59, 232}, { 68,  51, 179},
                        { 31,  40, 123}, { 13,  29,  77}, {  4,  17,  37},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {203,  41, 203}, {127,  56, 174}, { 49,  56, 127},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {110, 121, 217}, {119, 113, 213}, { 64,  95, 185},
                        { 30,  72, 144}, {  8,  42,  76}, {  2,  17,  25},
                    },
                    {  // band 2
                        {127, 159, 229}, {115, 134, 223}, { 36, 100, 189},
                        { 11,  75, 142}, {  3,  48,  83}, {  1,  19,  33},
                    },
                    {  // band 3
                        {150, 172, 241}, { 90, 133, 231}, { 28, 102, 192},
                        {  7,  81, 147}, {  1,  53,  91}, {  1,  25,  42},
                    },
                    {  // band 4
                        {184, 144, 248}, {114, 117, 237}, { 37,  89, 192},
                        { 10,  63, 130}, {  4,  42,  76}, {  1,  19,  38},
                    },
                    {  // band 5
                        {207,  79, 250}, {179,  74, 241}, { 83,  67, 199},
                        { 38,  51, 142}, { 17,  37,  97}, { 10,  14,  55},
                    },
                },
                {  // Inter
                    {  // band 0
                        {220,  82, 232}, {150,  93, 214}, { 66,  95, 177},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {116, 160, 227}, {136, 141, 227}, { 67, 114, 190},
                        { 40,  94, 148}, { 21,  70, 107}, { 10,  43,  63},
                    },
                    {  // band 2
                        {124, 173, 235}, {105, 147, 226}, { 27, 107, 184},
                        { 10,  80, 142}, {  3,  50,  86}, {  1,  16,  32},
                    },
                    {  // band 3
                        {149, 179, 243}, { 89, 147, 234}, { 29, 112, 193},
                        {  9,  94, 157}, {  1,  64, 111}, {  1,  25,  43},
                    },
                    {  // band 4
                        {187, 153, 248}, {127, 130, 241}, { 52,  99, 202},
                        { 20,  79, 152}, {  4,  50,  93}, {  1,  19,  32},
                    },
                    {  // band 5
                        {215,  82, 251}, {195,  80, 246}, { 93,  70, 204},
                        { 39,  54, 147}, { 14,  33,  88}, {  6,  14,  39},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {116,  43, 131}, { 39,  41,  94}, {  4,  28,  47},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 28, 101, 141}, { 27,  95, 140}, { 18,  80, 121},
                        { 10,  61,  95}, {  4,  39,  60}, {  1,  19,  26},
                    },
                    {  // band 2
                        { 29, 150, 183}, { 19, 127, 175}, {  8,  98, 147},
                        {  3,  76, 115}, {  1,  55,  84}, {  1,  29,  43},
                    },
                    {  // band 3
                        { 26, 168, 202}, { 12, 138, 188}, {  2,  98, 149},
                        {  1,  69, 110}, {  1,  40,  65}, {  1,  17,  25},
                    },
                    {  // band 4
                        { 33, 188, 225}, { 12, 155, 207}, {  2, 101, 155},
                        {  1,  65, 106}, {  1,  36,  60}, {  1,  18,  26},
                    },
                    {  // band 5
                        { 79, 205, 242}, { 30, 168, 224}, {  5, 106, 164},
                        {  1,  68, 110}, {  1,  39,  65}, {  1,  18,  28},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 96,  80, 201}, { 51,  88, 168}, { 14,  78, 116},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  6, 167, 216}, { 32, 152, 211}, { 24, 121, 182},
                        { 13,  98, 149}, { 12,  76, 108}, {  8,  48,  62},
                    },
                    {  // band 2
                        { 17, 176, 225}, { 13, 147, 209}, {  3,  96, 155},
                        {  1,  65, 108}, {  2,  43,  63}, {  2,  23,  25},
                    },
                    {  // band 3
                        { 18, 183, 232}, { 10, 153, 214}, {  1,  96, 154},
                        {  1,  63, 105}, {  1,  39,  59}, {  1,  21,  24},
                    },
                    {  // band 4
                        { 23, 191, 239}, {  8, 159, 221}, {  1,  97, 158},
                        {  1,  61, 105}, {  1,  37,  60}, {  1,  20,  26},
                    },
                    {  // band 5
                        { 70, 201, 243}, { 29, 163, 228}, {  4, 102, 169},
                        {  1,  67, 114}, {  1,  39,  66}, {  1,  17,  29},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {181,  38, 192}, { 95,  47, 151}, { 29,  49, 102},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 72, 131, 202}, { 93, 120, 205}, { 50, 103, 179},
                        { 24,  79, 143}, { 11,  47,  78}, {  7,  19,  25},
                    },
                    {  // band 2
                        { 84, 176, 221}, { 56, 144, 214}, { 21, 108, 182},
                        {  8,  83, 139}, {  3,  55,  90}, {  2,  27,  41},
                    },
                    {  // band 3
                        { 84, 195, 234}, { 42, 156, 222}, { 10, 109, 180},
                        {  4,  77, 133}, {  1,  48,  80}, {  1,  23,  35},
                    },
                    {  // band 4
                        { 89, 210, 238}, { 35, 165, 221}, {  6, 106, 172},
                        {  2,  70, 123}, {  1,  44,  74}, {  1,  21,  30},
                    },
                    {  // band 5
                        {114, 221, 247}, { 49, 170, 234}, {  7, 113, 184},
                        {  2,  77, 132}, {  1,  48,  79}, {  1,  25,  33},
                    },
                },
                {  // Inter
                    {  // band 0
                        {192,  66, 237}, {113,  84, 211}, { 35,  84, 154},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 81, 180, 234}, {127, 165, 229}, { 58, 137, 204},
                        { 41, 114, 174}, { 44,  94, 136}, { 29,  66,  86},
                    },
                    {  // band 2
                        { 82, 193, 240}, { 39, 162, 223}, {  8, 113, 179},
                        {  3,  83, 136}, {  6,  62,  84}, {  5,  45,  45},
                    },
                    {  // band 3
                        { 78, 203, 242}, { 31, 170, 227}, {  4, 115, 181},
                        {  1,  82, 135}, {  2,  59,  82}, {  1,  45,  47},
                    },
                    {  // band 4
                        { 76, 210, 239}, { 25, 170, 213}, {  2,  99, 152},
                        {  1,  69, 115}, {  1,  49,  80}, {  1,  47,  57},
                    },
                    {  // band 5
                        {103, 217, 250}, { 42, 180, 237}, {  3, 124, 191},
                        {  1,  90, 150}, {  1,  69, 116}, {  1,  52,  46},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 58,  38,  99}, {  9,  26,  51}, {  1,  14,  22},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 14,  78, 109}, { 16,  73, 105}, { 11,  62,  92},
                        {  6,  47,  72}, {  2,  29,  45}, {  1,  12,  18},
                    },
                    {  // band 2
                        { 17, 131, 148}, { 11, 112, 140}, {  5,  87, 118},
                        {  2,  63,  90}, {  1,  42,  63}, {  1,  19,  31},
                    },
                    {  // band 3
                        { 12, 151, 168}, {  6, 116, 152}, {  1,  76, 115},
                        {  1,  50,  81}, {  1,  32,  52}, {  1,  14,  23},
                    },
                    {  // band 4
                        { 10, 174, 191}, {  3, 130, 172}, {  1,  80, 126},
                        {  1,  53,  88}, {  1,  32,  55}, {  1,  14,  24},
                    },
                    {  // band 5
                        { 19, 219, 237}, {  3, 168, 211}, {  1,  90, 142},
                        {  1,  53,  91}, {  1,  29,  51}, {  1,  12,  21},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 21,  46, 184}, { 10,  53, 130}, {  2,  49,  78},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  3, 169, 198}, { 37, 165, 196}, { 26, 134, 176},
                        { 11, 108, 149}, {  5,  81, 112}, {  3,  47,  64},
                    },
                    {  // band 2
                        { 11, 183, 215}, {  8, 142, 192}, {  2,  91, 141},
                        {  1,  62, 100}, {  1,  38,  62}, {  1,  17,  28},
                    },
                    {  // band 3
                        { 12, 190, 223}, {  6, 149, 199}, {  1,  88, 139},
                        {  1,  56,  93}, {  1,  31,  54}, {  1,  13,  21},
                    },
                    {  // band 4
                        { 11, 197, 230}, {  3, 154, 204}, {  1,  83, 134},
                        {  1,  50,  86}, {  1,  28,  49}, {  1,  12,  21},
                    },
                    {  // band 5
                        { 17, 211, 240}, {  2, 167, 217}, {  1,  88, 143},
                        {  1,  53,  91}, {  1,  30,  53}, {  1,  14,  24},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {151,  30, 151}, { 50,  36, 105}, {  8,  34,  66},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 39, 111, 160}, { 62, 111, 165}, { 37,  99, 147},
                        { 15,  77, 118}, {  3,  47,  73}, {  1,  17,  27},
                    },
                    {  // band 2
                        { 48, 170, 190}, { 32, 135, 180}, { 11, 100, 149},
                        {  4,  76, 116}, {  1,  51,  80}, {  1,  22,  36},
                    },
                    {  // band 3
                        { 39, 191, 208}, { 18, 141, 191}, {  3,  96, 150},
                        {  1,  66, 110}, {  1,  41,  69}, {  1,  17,  28},
                    },
                    {  // band 4
                        { 32, 209, 219}, {  8, 152, 201}, {  1,  96, 153},
                        {  1,  63, 106}, {  1,  38,  66}, {  1,  17,  29},
                    },
                    {  // band 5
                        { 33, 230, 237}, {  5, 173, 214}, {  1, 100, 155},
                        {  1,  62, 105}, {  1,  38,  66}, {  1,  18,  32},
                    },
                },
                {  // Inter
                    {  // band 0
                        {149,  38, 231}, { 59,  51, 186}, { 12,  54, 117},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 53, 179, 226}, {126, 176, 223}, { 58, 147, 202},
                        { 28, 118, 174}, { 15,  94, 138}, { 14,  63,  87},
                    },
                    {  // band 2
                        { 58, 196, 232}, { 26, 158, 213}, {  5, 106, 166},
                        {  1,  75, 124}, {  1,  46,  79}, {  1,  23,  39},
                    },
                    {  // band 3
                        { 46, 203, 235}, { 17, 162, 213}, {  2, 104, 165},
                        {  1,  72, 120}, {  1,  44,  74}, {  1,  22,  33},
                    },
                    {  // band 4
                        { 37, 213, 238}, {  8, 167, 216}, {  1, 104, 168},
                        {  1,  68, 119}, {  1,  40,  67}, {  1,  17,  29},
                    },
                    {  // band 5
                        { 30, 228, 239}, {  4, 181, 213}, {  1, 103, 153},
                        {  1,  65, 110}, {  1,  43,  79}, {  1,  27,  56},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 76,  25,  53}, {  9,  18,  32}, {  1,  12,  18},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 29,  55,  91}, { 19,  58,  95}, { 15,  57,  89},
                        { 12,  49,  77}, {  3,  29,  44}, {  1,   8,  12},
                    },
                    {  // band 2
                        { 32, 160, 148}, { 33, 143, 146}, { 19, 122, 132},
                        {  6,  90, 102}, {  1,  58,  70}, {  1,  17,  24},
                    },
                    {  // band 3
                        { 16, 181, 181}, {  6, 142, 165}, {  1,  90, 120},
                        {  1,  50,  71}, {  1,  25,  38}, {  1,   9,  14},
                    },
                    {  // band 4
                        { 13, 203, 203}, {  3, 154, 176}, {  1,  80, 108},
                        {  1,  41,  61}, {  1,  24,  37}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 234, 240}, {  1, 178, 204}, {  1,  80, 119},
                        {  1,  45,  71}, {  1,  26,  42}, {  1,  12,  19},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 78,  20, 135}, { 25,  18, 101}, {  5,  19,  57},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  7, 144, 183}, {117, 151, 195}, {109, 151, 187},
                        { 39, 130, 168}, { 11, 100, 125}, {  4,  59,  64},
                    },
                    {  // band 2
                        { 20, 184, 212}, { 12, 148, 191}, {  2,  98, 141},
                        {  1,  65, 100}, {  1,  39,  61}, {  1,  14,  22},
                    },
                    {  // band 3
                        { 15, 194, 222}, {  6, 153, 198}, {  1,  92, 138},
                        {  1,  58,  91}, {  1,  32,  52}, {  1,  12,  18},
                    },
                    {  // band 4
                        { 14, 206, 232}, {  3, 162, 206}, {  1,  89, 134},
                        {  1,  52,  83}, {  1,  28,  46}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 225, 241}, {  1, 175, 210}, {  1,  81, 125},
                        {  1,  48,  78}, {  1,  28,  46}, {  1,  13,  21},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {124,  23,  93}, { 31,  24,  63}, {  6,  24,  46},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 23,  86, 126}, { 45,  90, 145}, { 31,  91, 133},
                        { 19,  80, 114}, {  7,  53,  72}, {  1,  20,  27},
                    },
                    {  // band 2
                        { 51, 186, 189}, { 48, 159, 182}, { 33, 128, 156},
                        { 15,  92, 124}, {  2,  62,  83}, {  1,  29,  43},
                    },
                    {  // band 3
                        { 36, 198, 211}, { 15, 156, 187}, {  3,  97, 137},
                        {  1,  61,  93}, {  1,  35,  57}, {  1,  15,  23},
                    },
                    {  // band 4
                        { 34, 219, 223}, {  9, 162, 193}, {  1,  91, 136},
                        {  1,  58,  92}, {  1,  35,  54}, {  1,  14,  23},
                    },
                    {  // band 5
                        { 19, 243, 243}, {  3, 191, 208}, {  1,  91, 137},
                        {  1,  56,  90}, {  1,  34,  55}, {  1,  16,  24},
                    },
                },
                {  // Inter
                    {  // band 0
                        {119,  20, 197}, { 19,  29, 156}, {  3,  30, 107},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 24, 192, 226}, {161, 193, 227}, { 97, 185, 222},
                        { 31, 158, 204}, { 16, 122, 165}, { 17,  84, 112},
                    },
                    {  // band 2
                        { 26, 202, 229}, { 11, 165, 210}, {  2, 103, 152},
                        {  1,  68, 104}, {  1,  42,  70}, {  1,  16,  36},
                    },
                    {  // band 3
                        { 24, 209, 237}, {  6, 169, 214}, {  1, 102, 154},
                        {  1,  65, 107}, {  1,  45,  68}, {  1,  17,  24},
                    },
                    {  // band 4
                        { 19, 219, 243}, {  4, 183, 226}, {  1, 115, 172},
                        {  1,  73, 119}, {  1,  43,  77}, {  1,  15,  37},
                    },
                    {  // band 5
                        { 11, 237, 241}, {  2, 190, 216}, {  1, 108, 146},
                        {  1,  59,  94}, {  1,  40,  67}, {  1,  30,  53},
                    },
                },
            },
        },
    },
    {  // Q_Index 1
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {174,  30, 159}, { 76,  38, 115}, { 15,  33,  65},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 60,  80, 153}, { 72,  75, 147}, { 36,  68, 129},
                        { 15,  59, 104}, {  4,  45,  74}, {  1,  28,  45},
                    },
                    {  // band 2
                        { 70, 122, 186}, { 55, 104, 175}, { 21,  83, 144},
                        {  8,  67, 112}, {  2,  51,  82}, {  1,  34,  57},
                    },
                    {  // band 3
                        { 97, 144, 207}, { 52, 109, 195}, { 16,  77, 153},
                        {  4,  58, 113}, {  1,  43,  77}, {  1,  27,  48},
                    },
                    {  // band 4
                        {128, 148, 229}, { 76, 104, 210}, { 18,  77, 159},
                        {  4,  65, 110}, {  1,  52,  82}, {  1,  31,  55},
                    },
                    {  // band 5
                        {165,  51, 238}, {128,  50, 230}, { 57,  49, 185},
                        { 28,  47, 130}, { 12,  44,  96}, {  3,  36,  60},
                    },
                },
                {  // Intra
                    {  // band 0
                        {169, 103, 203}, {117,  96, 176}, { 56,  81, 137},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 31, 150, 224}, { 49, 128, 212}, { 19,  92, 165},
                        {  6,  67, 116}, {  2,  43,  71}, {  1,  21,  36},
                    },
                    {  // band 2
                        { 58, 156, 230}, { 47, 130, 215}, {  7,  87, 158},
                        {  2,  63, 114}, {  1,  39,  71}, {  1,  18,  36},
                    },
                    {  // band 3
                        { 85, 176, 240}, { 43, 138, 226}, {  8,  93, 172},
                        {  2,  70, 127}, {  1,  46,  81}, {  1,  26,  47},
                    },
                    {  // band 4
                        {155, 144, 248}, { 93, 116, 235}, { 21,  83, 180},
                        {  4,  59, 119}, {  1,  43,  80}, {  1,  25,  50},
                    },
                    {  // band 5
                        {203,  61, 250}, {171,  57, 243}, { 71,  57, 199},
                        { 31,  49, 144}, { 13,  42,  96}, {  7,  30,  52},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {204,  44, 204}, {137,  57, 184}, { 72,  62, 152},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {145, 117, 236}, {151, 112, 231}, { 87,  95, 208},
                        { 31,  77, 165}, {  5,  49,  98}, {  1,  24,  39},
                    },
                    {  // band 2
                        {146, 152, 241}, {140, 132, 236}, { 41, 103, 209},
                        { 10,  86, 165}, {  2,  55, 106}, {  1,  25,  58},
                    },
                    {  // band 3
                        {154, 181, 249}, { 84, 143, 240}, { 23, 114, 210},
                        {  6, 102, 182}, {  2,  71, 137}, {  1,  35,  90},
                    },
                    {  // band 4
                        {184, 150, 251}, {115, 130, 244}, { 34, 105, 215},
                        { 15,  89, 173}, {  1,  51, 141}, {128, 128, 128},
                    },
                    {  // band 5
                        {211,  71, 253}, {193,  78, 249}, {106,  91, 232},
                        { 61,  87, 198}, { 85, 153, 254}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {232, 104, 242}, {165, 114, 227}, { 96, 120, 206},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {137, 178, 250}, {146, 153, 245}, { 74, 108, 205},
                        { 41,  81, 149}, { 24,  55, 104}, { 13,  36,  68},
                    },
                    {  // band 2
                        {147, 185, 252}, {127, 161, 246}, { 30, 104, 208},
                        { 11,  74, 154}, {  6,  54, 100}, {  2,  29,  63},
                    },
                    {  // band 3
                        {163, 191, 254}, {101, 161, 249}, { 22, 114, 215},
                        {  6,  89, 173}, {  1,  65, 120}, {  1,   1, 170},
                    },
                    {  // band 4
                        {197, 160, 254}, {142, 141, 251}, { 39, 102, 218},
                        { 10,  76, 158}, {  1,  56, 122}, {128, 128, 128},
                    },
                    {  // band 5
                        {224,  76, 254}, {215,  84, 253}, {107,  85, 232},
                        { 43,  71, 177}, {  1,   1, 254}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 68,  37, 120}, { 21,  34,  82}, {  5,  26,  49},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 41,  89, 138}, { 56,  83, 132}, { 31,  73, 115},
                        { 16,  62,  92}, {  5,  45,  62}, {  1,  24,  32},
                    },
                    {  // band 2
                        { 48, 139, 165}, { 30, 114, 160}, { 13,  92, 132},
                        {  6,  72, 103}, {  3,  49,  72}, {  1,  26,  41},
                    },
                    {  // band 3
                        { 44, 162, 191}, { 20, 127, 175}, {  5,  90, 137},
                        {  1,  62, 100}, {  1,  38,  63}, {  1,  20,  32},
                    },
                    {  // band 4
                        { 51, 184, 213}, { 16, 137, 193}, {  2,  89, 143},
                        {  1,  60, 102}, {  1,  39,  66}, {  1,  23,  37},
                    },
                    {  // band 5
                        { 76, 200, 235}, { 27, 150, 216}, {  3,  99, 164},
                        {  1,  70, 119}, {  1,  45,  77}, {  1,  22,  38},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 81, 112, 199}, { 49, 101, 164}, { 19,  80, 119},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 12, 181, 217}, { 48, 151, 212}, { 38, 118, 180},
                        { 22,  95, 140}, { 11,  67,  92}, { 13,  46,  44},
                    },
                    {  // band 2
                        { 29, 188, 226}, { 19, 147, 210}, {  5,  95, 154},
                        {  4,  68, 106}, {  3,  44,  60}, {  1,  24,  27},
                    },
                    {  // band 3
                        { 30, 195, 234}, { 15, 153, 216}, {  3,  95, 156},
                        {  2,  66, 108}, {  2,  44,  62}, {  1,  24,  29},
                    },
                    {  // band 4
                        { 36, 203, 243}, { 12, 162, 225}, {  2,  98, 163},
                        {  2,  67, 113}, {  2,  45,  68}, {  1,  24,  34},
                    },
                    {  // band 5
                        { 86, 207, 248}, { 35, 165, 236}, {  3, 107, 180},
                        {  1,  73, 128}, {  1,  45,  78}, {  1,  20,  34},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {188,  37, 205}, {118,  51, 172}, { 56,  57, 135},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {116, 135, 225}, {144, 123, 221}, { 72, 103, 197},
                        { 35,  77, 153}, { 15,  47,  82}, {  6,  25,  34},
                    },
                    {  // band 2
                        {128, 171, 233}, { 82, 142, 226}, { 31, 106, 191},
                        { 16,  82, 146}, {  9,  59,  98}, {  4,  33,  54},
                    },
                    {  // band 3
                        {126, 197, 241}, { 66, 155, 230}, { 18, 108, 190},
                        {  7,  82, 148}, {  3,  58,  98}, {  1,  25,  50},
                    },
                    {  // band 4
                        {117, 207, 244}, { 44, 163, 233}, {  9, 112, 191},
                        {  5,  84, 148}, {  3,  61,  87}, {  1,  28,  38},
                    },
                    {  // band 5
                        {112, 214, 249}, { 39, 174, 240}, {  6, 125, 205},
                        {  4,  96, 163}, {  5,  66, 100}, {  1, 128, 254},
                    },
                },
                {  // Inter
                    {  // band 0
                        {227,  70, 234}, {145,  91, 213}, { 61, 100, 173},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {108, 198, 243}, {171, 172, 240}, {118, 130, 210},
                        {104, 107, 165}, { 64,  85, 114}, { 55,  64,  60},
                    },
                    {  // band 2
                        {110, 208, 247}, { 64, 175, 237}, { 24, 112, 187},
                        { 24,  81, 133}, { 24,  63,  83}, { 21,  47,  53},
                    },
                    {  // band 3
                        { 91, 218, 249}, { 46, 188, 238}, {  8, 113, 184},
                        {  5,  83, 137}, {  6,  62,  95}, { 17,  44,  94},
                    },
                    {  // band 4
                        { 84, 216, 248}, { 30, 187, 237}, {  2, 117, 188},
                        {  1,  88, 141}, {  3,  63,  98}, {  1,   1,   1},
                    },
                    {  // band 5
                        {116, 218, 252}, { 47, 186, 242}, {  2, 132, 204},
                        {  1, 106, 175}, {  1,  88, 104}, {  1, 254, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 35,  41, 129}, { 12,  30,  70}, {  2,  19,  32},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 30,  77, 116}, { 39,  70, 110}, { 20,  58,  96},
                        {  8,  47,  77}, {  2,  33,  52}, {  1,  17,  26},
                    },
                    {  // band 2
                        { 31, 123, 146}, { 18, 103, 140}, {  7,  81, 119},
                        {  2,  62,  95}, {  1,  44,  70}, {  1,  26,  42},
                    },
                    {  // band 3
                        { 21, 149, 170}, {  9, 114, 158}, {  2,  80, 126},
                        {  1,  57,  94}, {  1,  36,  61}, {  1,  18,  31},
                    },
                    {  // band 4
                        { 20, 178, 199}, {  6, 134, 183}, {  1,  87, 139},
                        {  1,  60, 100}, {  1,  37,  64}, {  1,  18,  31},
                    },
                    {  // band 5
                        { 36, 218, 233}, {  6, 160, 207}, {  1,  92, 147},
                        {  1,  59, 101}, {  1,  35,  62}, {  1,  18,  31},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 17,  62, 211}, { 14,  62, 153}, {  5,  50,  84},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 11, 180, 205}, { 87, 160, 205}, { 53, 128, 184},
                        { 27, 106, 156}, { 13,  79, 115}, {  6,  46,  67},
                    },
                    {  // band 2
                        { 32, 194, 220}, { 20, 145, 202}, {  4,  96, 152},
                        {  1,  67, 111}, {  1,  42,  70}, {  1,  21,  37},
                    },
                    {  // band 3
                        { 30, 204, 228}, { 14, 152, 207}, {  1,  92, 149},
                        {  1,  61, 103}, {  1,  34,  59}, {  1,  16,  28},
                    },
                    {  // band 4
                        { 27, 213, 235}, {  7, 159, 210}, {  1,  88, 143},
                        {  1,  55,  94}, {  1,  31,  53}, {  1,  16,  27},
                    },
                    {  // band 5
                        { 28, 223, 243}, {  4, 173, 217}, {  1,  91, 146},
                        {  1,  58,  98}, {  1,  35,  60}, {  1,  19,  33},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {172,  37, 202}, { 83,  51, 156}, { 24,  53, 110},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 76, 134, 206}, {110, 124, 200}, { 47, 106, 180},
                        { 15,  82, 145}, {  3,  48,  83}, {  1,  19,  32},
                    },
                    {  // band 2
                        { 80, 176, 220}, { 49, 145, 212}, { 17, 112, 180},
                        {  7,  84, 140}, {  1,  53,  89}, {  1,  27,  43},
                    },
                    {  // band 3
                        { 74, 201, 232}, { 38, 158, 221}, {  8, 112, 179},
                        {  2,  79, 132}, {  1,  47,  82}, {  1,  26,  42},
                    },
                    {  // band 4
                        { 73, 215, 239}, { 28, 169, 227}, {  3, 112, 176},
                        {  1,  74, 126}, {  1,  48,  79}, {  1,  27,  44},
                    },
                    {  // band 5
                        { 71, 233, 244}, { 18, 180, 230}, {  1, 114, 180},
                        {  1,  80, 134}, {  1,  51,  85}, {  1,  26,  36},
                    },
                },
                {  // Inter
                    {  // band 0
                        {213,  34, 244}, {126,  57, 212}, { 46,  67, 151},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {120, 202, 245}, {198, 173, 241}, {119, 146, 224},
                        { 76, 126, 195}, { 44, 102, 159}, { 40,  76, 115},
                    },
                    {  // band 2
                        {120, 215, 248}, { 69, 171, 237}, { 23, 119, 194},
                        { 10,  86, 147}, {  2,  56,  94}, {  1,  25,  44},
                    },
                    {  // band 3
                        {102, 226, 250}, { 53, 183, 239}, {  9, 118, 188},
                        {  2,  78, 131}, {  1,  48,  89}, {  1,  17,  36},
                    },
                    {  // band 4
                        { 86, 235, 252}, { 34, 194, 240}, {  2, 109, 173},
                        {  1,  68, 118}, {  1,  44,  79}, {  1,   1,  38},
                    },
                    {  // band 5
                        { 59, 236, 243}, { 11, 189, 228}, {  1, 112, 187},
                        {  1,  88, 145}, {  1,  55,  92}, {  1,   1, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 41,  40, 104}, { 12,  31,  64}, {  2,  16,  28},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 65,  58, 132}, { 50,  61, 130}, { 40,  57, 116},
                        { 22,  46,  87}, {  2,  28,  44}, {  1,  11,  17},
                    },
                    {  // band 2
                        { 55, 139, 135}, { 46, 122, 132}, { 21,  89, 110},
                        {  6,  60,  78}, {  1,  38,  54}, {  1,  17,  27},
                    },
                    {  // band 3
                        { 29, 167, 161}, { 10, 120, 141}, {  1,  69,  98},
                        {  1,  42,  66}, {  1,  28,  44}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 19, 191, 180}, {  4, 125, 154}, {  1,  70, 107},
                        {  1,  48,  77}, {  1,  33,  53}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 16, 238, 231}, {  2, 163, 198}, {  1,  85, 134},
                        {  1,  54,  90}, {  1,  34,  57}, {  1,  17,  29},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 70,  15, 216}, { 40,  18, 164}, { 14,  17,  83},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 25, 150, 200}, {185, 154, 211}, {123, 137, 199},
                        { 67, 119, 177}, { 31,  96, 137}, { 18,  63,  86},
                    },
                    {  // band 2
                        { 57, 187, 223}, { 35, 148, 207}, {  7, 104, 159},
                        {  2,  72, 113}, {  1,  44,  71}, {  1,  20,  34},
                    },
                    {  // band 3
                        { 44, 203, 233}, { 18, 157, 212}, {  1,  98, 150},
                        {  1,  61, 102}, {  1,  38,  62}, {  1,  19,  31},
                    },
                    {  // band 4
                        { 41, 215, 238}, { 11, 166, 215}, {  1,  94, 146},
                        {  1,  60, 101}, {  1,  37,  63}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 19, 236, 246}, {  3, 188, 223}, {  1,  95, 146},
                        {  1,  58,  95}, {  1,  34,  56}, {  1,  17,  27},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {146,  27, 156}, { 49,  32, 116}, { 10,  39,  77},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 47, 101, 172}, { 93, 100, 178}, { 58,  91, 165},
                        { 26,  75, 134}, {  4,  49,  82}, {  2,  22,  33},
                    },
                    {  // band 2
                        { 60, 158, 196}, { 44, 135, 186}, { 25, 106, 157},
                        {  8,  81, 124}, {  2,  56,  86}, {  1,  28,  45},
                    },
                    {  // band 3
                        { 44, 169, 212}, { 15, 138, 196}, {  2, 100, 157},
                        {  1,  74, 119}, {  1,  49,  76}, {  1,  20,  34},
                    },
                    {  // band 4
                        { 38, 199, 231}, { 11, 158, 214}, {  1, 111, 167},
                        {  1,  76, 122}, {  1,  44,  76}, {  1,  17,  39},
                    },
                    {  // band 5
                        { 40, 236, 246}, { 10, 187, 230}, {  1, 115, 175},
                        {  1,  74, 122}, {  1,  42,  71}, {  1,  14,  59},
                    },
                },
                {  // Inter
                    {  // band 0
                        {161,  26, 237}, { 65,  46, 209}, { 21,  46, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 87, 229, 245}, {206, 214, 244}, {148, 186, 236},
                        { 89, 165, 221}, { 41, 132, 186}, { 37,  93, 141},
                    },
                    {  // band 2
                        { 93, 231, 246}, { 47, 181, 231}, {  8, 117, 188},
                        {  2,  84, 138}, {  1,  43,  87}, {  1,  27,  41},
                    },
                    {  // band 3
                        { 80, 239, 250}, { 28, 190, 236}, {  1, 119, 183},
                        {  1,  84, 135}, {  1,  81,  69}, {  1, 102,   1},
                    },
                    {  // band 4
                        { 67, 245, 252}, { 22, 206, 242}, {  1, 130, 195},
                        {  1,  77, 136}, {  1,  35,  88}, {128, 128, 128},
                    },
                    {  // band 5
                        { 43, 250, 228}, { 31, 185, 204}, {  6, 101, 183},
                        {  1,  92, 151}, {  1,  84, 137}, {128, 128, 128},
                    },
                },
            },
        },
    },
    {  // Q_Index 2
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {181,  22, 175}, { 96,  37, 147}, { 35,  41, 105},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80,  95, 197}, {111,  92, 193}, { 59,  87, 175},
                        { 29,  79, 150}, { 10,  65, 118}, {  2,  47,  82},
                    },
                    {  // band 2
                        { 90, 141, 216}, { 77, 120, 210}, { 23,  95, 184},
                        { 11,  81, 151}, {  6,  75, 130}, {  2,  58, 113},
                    },
                    {  // band 3
                        {122, 167, 231}, { 66, 119, 225}, { 26,  87, 189},
                        {  7,  76, 151}, {  2,  63, 125}, {  1,  59,  77},
                    },
                    {  // band 4
                        {162, 147, 244}, {110,  97, 236}, { 32,  88, 204},
                        { 11,  89, 174}, {  5,  78, 151}, {128, 128, 128},
                    },
                    {  // band 5
                        {205,  59, 251}, {176,  68, 248}, { 90,  71, 223},
                        { 49,  72, 188}, { 17,  74, 203}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {188,  70, 207}, {140,  73, 189}, { 85,  73, 163},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 59, 144, 239}, { 79, 126, 237}, { 31, 102, 202},
                        { 10,  81, 153}, {  3,  56, 102}, {  2,  33,  59},
                    },
                    {  // band 2
                        {100, 152, 243}, { 80, 129, 236}, { 14,  94, 194},
                        {  4,  72, 150}, {  1,  50, 103}, {  1,  35,  60},
                    },
                    {  // band 3
                        {130, 183, 247}, { 70, 139, 242}, { 19, 100, 203},
                        {  4,  83, 159}, {  1,  59, 119}, {  1,  44,  72},
                    },
                    {  // band 4
                        {197, 138, 252}, {135, 107, 247}, { 31,  86, 210},
                        {  7,  74, 160}, {  1,  53, 107}, {128, 128, 128},
                    },
                    {  // band 5
                        {229,  54, 254}, {200,  51, 251}, { 83,  61, 226},
                        { 33,  55, 177}, { 12,  74, 145}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {229,  20, 235}, {183,  37, 221}, {127,  47, 198},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {188, 115, 251}, {208, 110, 250}, {101,  99, 235},
                        { 38,  81, 197}, {  9,  56, 132}, {  9,  52,  63},
                    },
                    {  // band 2
                        {189, 150, 252}, {186, 137, 251}, { 54, 107, 236},
                        { 14,  90, 195}, {  1,  89, 104}, {128, 128, 128},
                    },
                    {  // band 3
                        {209, 180, 254}, {142, 145, 253}, { 51, 130, 236},
                        {  6, 128, 214}, {  1, 128, 254}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 140, 254}, {194, 128, 254}, { 75, 119, 233},
                        {128,  23, 230}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {244,  59, 254}, {239,  81, 254}, {128,  85, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {246,  55, 247}, {197,  64, 235}, {141,  74, 218},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {178, 163, 254}, {192, 138, 252}, { 85, 103, 231},
                        { 49,  81, 179}, { 32,  54, 133}, { 12,  26,  98},
                    },
                    {  // band 2
                        {189, 173, 254}, {179, 150, 253}, { 60,  94, 237},
                        { 34,  81, 198}, { 20,  53, 187}, {128, 128, 128},
                    },
                    {  // band 3
                        {202, 191, 254}, {157, 160, 254}, { 57, 117, 240},
                        { 28, 105, 211}, {  1, 128,   1}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 146, 254}, {208, 133, 254}, { 66,  78, 233},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  49, 254}, {246,  63, 254}, { 85, 142, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 45,  28, 124}, { 23,  35, 107}, { 10,  34,  78},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 53,  99, 177}, { 82,  96, 174}, { 46,  89, 158},
                        { 21,  76, 133}, {  6,  56,  94}, {  1,  33,  54},
                    },
                    {  // band 2
                        { 68, 147, 201}, { 42, 124, 195}, { 17,  98, 166},
                        {  7,  75, 131}, {  2,  53,  93}, {  1,  33,  59},
                    },
                    {  // band 3
                        { 65, 176, 217}, { 30, 137, 206}, {  6,  97, 167},
                        {  2,  70, 128}, {  1,  47,  88}, {  1,  29,  46},
                    },
                    {  // band 4
                        { 69, 195, 232}, { 24, 146, 218}, {  4, 100, 175},
                        {  2,  72, 134}, {  1,  51,  93}, {  1,  29,  52},
                    },
                    {  // band 5
                        { 96, 212, 246}, { 39, 158, 234}, {  6, 109, 192},
                        {  2,  77, 144}, {  1,  50,  95}, {  1,  20,  45},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 71,  80, 213}, { 53,  73, 181}, { 25,  66, 141},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35, 168, 231}, { 91, 150, 229}, { 49, 122, 202},
                        { 22,  97, 162}, { 10,  68, 108}, {  9,  48,  57},
                    },
                    {  // band 2
                        { 56, 178, 236}, { 32, 148, 225}, {  9,  99, 176},
                        {  4,  69, 127}, {  2,  44,  78}, {  1,  25,  41},
                    },
                    {  // band 3
                        { 57, 191, 242}, { 27, 155, 230}, {  5, 102, 180},
                        {  2,  71, 133}, {  1,  44,  78}, {  1,  27,  41},
                    },
                    {  // band 4
                        { 67, 201, 247}, { 24, 162, 237}, {  3, 106, 188},
                        {  3,  74, 137}, {  1,  46,  85}, {  1,  34,  48},
                    },
                    {  // band 5
                        {111, 210, 251}, { 47, 166, 244}, {  3, 113, 199},
                        {  2,  77, 146}, {  1,  48,  93}, {  1,  38,  22},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {206,  21, 221}, {150,  36, 195}, { 94,  44, 164},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {147, 128, 239}, {194, 122, 238}, { 95, 104, 220},
                        { 39,  81, 183}, { 13,  53, 111}, {  3,  24,  49},
                    },
                    {  // band 2
                        {164, 163, 244}, {106, 142, 239}, { 50, 112, 215},
                        { 26,  90, 177}, { 12,  67, 130}, {  1,   1,  64},
                    },
                    {  // band 3
                        {155, 193, 249}, { 88, 158, 244}, { 26, 124, 220},
                        { 10,  98, 173}, {  1,  77, 126}, {128, 128, 128},
                    },
                    {  // band 4
                        {141, 205, 252}, { 64, 174, 248}, { 17, 124, 221},
                        { 12,  92, 176}, {  1,  29, 148}, {128, 128, 128},
                    },
                    {  // band 5
                        {150, 217, 254}, { 74, 191, 252}, { 30, 144, 215},
                        {  1, 106, 137}, {128,   1, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {241,  37, 242}, {175,  48, 223}, { 99,  53, 189},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {153, 183, 248}, {212, 156, 247}, {134, 124, 221},
                        { 88, 103, 184}, { 59,  86, 132}, { 29,  61,  67},
                    },
                    {  // band 2
                        {162, 199, 250}, {106, 167, 247}, { 56, 110, 207},
                        { 32,  85, 165}, { 16,  71, 130}, {  1,  93, 254},
                    },
                    {  // band 3
                        {143, 213, 252}, { 86, 187, 250}, { 23, 124, 220},
                        {  7,  95, 176}, {  1, 109, 102}, {128, 128, 128},
                    },
                    {  // band 4
                        {130, 219, 254}, { 70, 201, 253}, { 15, 128, 215},
                        {  1, 101, 201}, {  1,  64, 170}, {128, 128, 128},
                    },
                    {  // band 5
                        {155, 219, 254}, {105, 207, 254}, { 28, 155, 229},
                        {  1, 153, 191}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 18,  26, 117}, { 10,  29,  82}, {  3,  25,  52},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35,  88, 152}, { 62,  85, 150}, { 36,  77, 137},
                        { 16,  66, 116}, {  4,  47,  81}, {  1,  26,  44},
                    },
                    {  // band 2
                        { 55, 141, 182}, { 32, 119, 177}, { 12,  93, 154},
                        {  4,  71, 123}, {  1,  51,  89}, {  1,  32,  56},
                    },
                    {  // band 3
                        { 46, 171, 202}, { 21, 130, 191}, {  5,  91, 154},
                        {  1,  64, 115}, {  1,  42,  77}, {  1,  25,  41},
                    },
                    {  // band 4
                        { 43, 195, 219}, { 12, 142, 203}, {  1,  91, 156},
                        {  1,  63, 115}, {  1,  41,  77}, {  1,  22,  43},
                    },
                    {  // band 5
                        { 42, 221, 238}, {  8, 162, 219}, {  1,  98, 167},
                        {  1,  67, 123}, {  1,  43,  83}, {  1,  25,  38},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 16,  51, 216}, { 20,  48, 168}, {  9,  44, 109},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 34, 164, 226}, {124, 148, 226}, { 72, 127, 207},
                        { 36, 107, 175}, { 15,  81, 129}, {  6,  51,  79},
                    },
                    {  // band 2
                        { 61, 182, 234}, { 35, 148, 220}, {  9, 101, 178},
                        {  4,  71, 134}, {  1,  46,  90}, {  1,  24,  51},
                    },
                    {  // band 3
                        { 54, 198, 239}, { 25, 156, 224}, {  3,  98, 173},
                        {  1,  66, 124}, {  1,  41,  78}, {  1,  15,  37},
                    },
                    {  // band 4
                        { 48, 209, 242}, { 12, 162, 226}, {  1,  96, 169},
                        {  1,  63, 119}, {  1,  40,  78}, {  1,  18,  45},
                    },
                    {  // band 5
                        { 44, 223, 247}, {  6, 173, 232}, {  1, 105, 178},
                        {  1,  71, 131}, {  1,  44,  84}, {  1,  13,  46},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {188,  26, 214}, {121,  42, 181}, { 66,  49, 149},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {136, 128, 233}, {172, 124, 230}, { 80, 106, 211},
                        { 27,  81, 174}, {  6,  49,  98}, {  8,  28,  49},
                    },
                    {  // band 2
                        {145, 166, 239}, { 92, 141, 229}, { 28, 108, 196},
                        {  8,  87, 154}, {  1,  58, 105}, {  1,  27,  59},
                    },
                    {  // band 3
                        {131, 193, 242}, { 66, 151, 231}, { 13, 112, 192},
                        {  2,  81, 152}, {  1,  66, 121}, {  1,  23,  64},
                    },
                    {  // band 4
                        {112, 211, 246}, { 41, 164, 235}, {  5, 117, 202},
                        {  1,  83, 162}, {  1,  64, 111}, {128, 128, 128},
                    },
                    {  // band 5
                        { 96, 230, 250}, { 28, 185, 243}, {  2, 132, 204},
                        {  1,  91, 166}, {  1,  85,  46}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {238,  23, 242}, {157,  29, 215}, { 73,  27, 162},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {165, 173, 250}, {222, 151, 247}, {152, 134, 235},
                        {114, 120, 210}, { 86, 109, 176}, { 53,  88, 145},
                    },
                    {  // band 2
                        {164, 194, 249}, {100, 158, 241}, { 35, 111, 212},
                        { 17,  85, 167}, {  1,  52, 112}, {  1,  73,   1},
                    },
                    {  // band 3
                        {151, 215, 252}, { 83, 172, 245}, { 16, 122, 208},
                        {  6, 101, 165}, {  1,  74, 113}, {  1,   1,   1},
                    },
                    {  // band 4
                        {138, 230, 253}, { 65, 184, 248}, {  8, 128, 212},
                        {  1, 111, 182}, {128,   1,   1}, {128, 128, 128},
                    },
                    {  // band 5
                        {123, 240, 253}, { 36, 201, 250}, {  3, 127, 211},
                        {  1,  68, 204}, {128,   1,   1}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 51,  21, 156}, { 30,  23,  86}, {  4,  18,  37},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 38,  77, 129}, { 79,  76, 129}, { 40,  66, 117},
                        { 12,  54,  95}, {  1,  36,  60}, {  1,  17,  29},
                    },
                    {  // band 2
                        { 44, 133, 149}, { 24, 107, 143}, {  8,  78, 121},
                        {  3,  59,  97}, {  1,  42,  71}, {  1,  22,  37},
                    },
                    {  // band 3
                        { 29, 160, 171}, {  9, 114, 158}, {  1,  76, 125},
                        {  1,  54,  93}, {  1,  36,  63}, {  1,  20,  35},
                    },
                    {  // band 4
                        { 22, 188, 205}, {  6, 132, 186}, {  1,  87, 144},
                        {  1,  62, 107}, {  1,  41,  72}, {  1,  23,  41},
                    },
                    {  // band 5
                        { 25, 233, 236}, {  5, 165, 214}, {  1,  96, 158},
                        {  1,  63, 112}, {  1,  40,  73}, {  1,  23,  40},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 48,  20, 231}, { 37,  21, 179}, { 15,  18, 109},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 41, 154, 216}, {196, 142, 221}, {131, 125, 207},
                        { 84, 111, 181}, { 45,  91, 142}, { 27,  62,  89},
                    },
                    {  // band 2
                        { 72, 181, 230}, { 41, 147, 215}, { 10, 102, 173},
                        {  3,  73, 132}, {  1,  47,  89}, {  1,  23,  50},
                    },
                    {  // band 3
                        { 60, 201, 236}, { 23, 157, 219}, {  2,  99, 167},
                        {  1,  69, 124}, {  1,  43,  80}, {  1,  22,  39},
                    },
                    {  // band 4
                        { 53, 214, 242}, { 15, 165, 224}, {  1, 101, 173},
                        {  1,  70, 131}, {  1,  44,  83}, {  1,  23,  49},
                    },
                    {  // band 5
                        { 39, 239, 248}, {  7, 186, 233}, {  1, 108, 174},
                        {  1,  70, 123}, {  1,  43,  77}, {  1,  16,  42},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {161,  26, 204}, { 77,  40, 160}, { 26,  50, 117},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80, 140, 218}, {136, 133, 215}, { 63, 117, 197},
                        { 20,  93, 170}, {  7,  55, 102}, { 13,  32,  52},
                    },
                    {  // band 2
                        { 86, 173, 231}, { 46, 150, 220}, { 18, 118, 190},
                        {  8,  90, 150}, {  2,  60,  95}, {  1,  39,  41},
                    },
                    {  // band 3
                        { 80, 183, 242}, { 37, 160, 231}, {  6, 120, 182},
                        {  1,  86, 137}, {  1,  46,  78}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 88, 215, 247}, { 42, 179, 235}, {  4, 116, 182},
                        {  2,  80, 133}, {  1,  46,  85}, {  1,  64,  43},
                    },
                    {  // band 5
                        {100, 236, 250}, { 31, 186, 234}, {  1, 114, 181},
                        {  1,  85, 135}, {  1,  78,  64}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {213,  13, 245}, {106,  16, 211}, { 32,  11, 156},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {140, 214, 247}, {241, 186, 243}, {177, 172, 235},
                        {128, 156, 219}, {106, 130, 191}, { 99, 105, 152},
                    },
                    {  // band 2
                        {125, 218, 248}, { 75, 167, 239}, { 29, 111, 212},
                        {  6,  66, 152}, {  1,  42,  96}, {  1,  85, 128},
                    },
                    {  // band 3
                        {120, 232, 252}, { 60, 189, 247}, {  8, 141, 200},
                        {  1,  89, 134}, {  1,  32, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {111, 238, 253}, { 56, 198, 245}, {  1, 123, 208},
                        {  1,  93, 176}, {  1,   1,  73}, {128, 128, 128},
                    },
                    {  // band 5
                        { 98, 251, 249}, { 56, 189, 244}, { 17, 113, 220},
                        {  1, 109, 179}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
    },
    {  // Q_Index 3
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {186,  16, 200}, {122,  31, 187}, { 78,  40, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {146, 119, 245}, {182, 115, 244}, {130, 113, 238},
                        { 88, 110, 225}, { 47, 103, 208}, {  5, 102, 188},
                    },
                    {  // band 2
                        {164, 157, 248}, {155, 141, 250}, { 71, 116, 243},
                        { 88, 129, 233}, { 50,  99, 228}, { 26, 148, 191},
                    },
                    {  // band 3
                        {200, 158, 253}, {177, 118, 252}, { 99, 113, 245},
                        { 77, 120, 210}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 104, 254}, {209,  82, 254}, {143, 112, 252},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {250,  36, 254}, {243,  55, 254}, {223, 170, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {207,  37, 226}, {164,  46, 218}, {122,  58, 201},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {149, 154, 253}, {170, 137, 253}, { 94, 123, 247},
                        { 42, 113, 222}, { 16,  97, 174}, { 49,  98, 159},
                    },
                    {  // band 2
                        {177, 162, 253}, {165, 142, 252}, { 51, 108, 243},
                        { 18, 108, 213}, {  1,  98, 254}, {128, 128, 128},
                    },
                    {  // band 3
                        {211, 152, 254}, {184, 116, 254}, { 70, 110, 244},
                        {  8, 108, 237}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {236,  89, 254}, {210,  67, 254}, {112, 111, 248},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  26, 254}, {233,  35, 254}, {128,   1, 254},
                        {254, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {247,   2, 247}, {226,   8, 242}, {191,  14, 235},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {231,  94, 254}, {248,  91, 254}, {186,  89, 252},
                        {128,  92, 244}, { 79, 112, 254}, {128, 128, 128},
                    },
                    {  // band 2
                        {228, 145, 253}, {240, 130, 254}, {223, 105, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {245, 153, 253}, {240, 120, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 128, 254}, {204, 128, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {253,   7, 249}, {224,   9, 244}, {182,  13, 231},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {234, 109, 254}, {242, 104, 254}, {160,  98, 254},
                        {123,  85, 243}, { 82,  43, 217}, {128, 128, 128},
                    },
                    {  // band 2
                        {243, 137, 254}, {240, 118, 254}, {136,  53, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {251, 173, 254}, {229, 129, 250}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 119, 254}, {254, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 49,  26, 159}, { 36,  34, 150}, { 26,  38, 124},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 99, 122, 226}, {143, 119, 225}, { 90, 113, 213},
                        { 46, 102, 193}, { 14,  84, 157}, {  3,  59, 107},
                    },
                    {  // band 2
                        {109, 164, 237}, { 74, 142, 233}, { 29, 112, 216},
                        { 14,  92, 184}, { 10,  80, 156}, {  1,  52, 137},
                    },
                    {  // band 3
                        {110, 191, 245}, { 59, 156, 240}, { 18, 121, 220},
                        {  8,  97, 184}, {  3,  84, 150}, {128, 128, 128},
                    },
                    {  // band 4
                        {115, 203, 250}, { 59, 167, 246}, { 16, 130, 226},
                        {  7,  97, 192}, {  1,  71,  99}, {128, 128, 128},
                    },
                    {  // band 5
                        {149, 218, 253}, { 93, 171, 251}, { 28, 125, 233},
                        { 28,  99, 192}, {128,  85,  85}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 97,  45, 229}, { 79,  52, 205}, { 46,  58, 171},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 99, 180, 249}, {156, 165, 249}, { 73, 141, 237},
                        { 31, 116, 208}, { 13,  81, 153}, {  5,  42,  86},
                    },
                    {  // band 2
                        {113, 188, 251}, { 68, 161, 244}, { 16, 108, 216},
                        {  6,  81, 168}, {  2,  65, 118}, {128,   1,   1},
                    },
                    {  // band 3
                        {117, 201, 252}, { 62, 171, 248}, { 12, 119, 221},
                        {  5,  90, 182}, {  4,  66, 116}, {128, 128, 128},
                    },
                    {  // band 4
                        {128, 207, 253}, { 70, 176, 251}, { 11, 126, 228},
                        {  6,  89, 189}, {  1,  44, 148}, {128, 128, 128},
                    },
                    {  // band 5
                        {162, 218, 254}, {107, 170, 253}, { 22, 131, 238},
                        {  1,  77, 182}, {  1, 254, 128}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {235,   5, 238}, {194,  14, 223}, {152,  22, 205},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {200, 121, 251}, {241, 115, 252}, {167, 108, 248},
                        { 93,  93, 233}, { 36,  66, 189}, {128, 128, 128},
                    },
                    {  // band 2
                        {220, 151, 253}, {176, 135, 252}, { 95, 124, 254},
                        { 64, 105, 217}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {225, 189, 254}, {175, 155, 254}, {102, 119, 254},
                        {  1,   1,   1}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {218, 195, 254}, {125, 157, 253}, {128, 128, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {221, 197, 254}, { 85, 210, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {250,   9, 246}, {204,  13, 234}, {144,  18, 211},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {213, 157, 253}, {243, 138, 253}, {170, 117, 250},
                        {109,  91, 233}, { 66,  77, 163}, { 64,  85, 254},
                    },
                    {  // band 2
                        {221, 169, 254}, {182, 141, 253}, {112, 120, 239},
                        { 85, 165, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {226, 192, 254}, {189, 174, 251}, {153, 128, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {232, 192, 254}, {195, 187, 247}, {  1, 191, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {247, 185, 254}, {254,  93, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 14,  30, 136}, { 15,  33, 120}, { 10,  33,  90},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 92, 109, 209}, {113, 108, 207}, { 77, 102, 193},
                        { 39,  91, 171}, { 11,  70, 129}, {  2,  44,  77},
                    },
                    {  // band 2
                        { 99, 158, 223}, { 66, 135, 217}, { 23, 109, 194},
                        {  9,  85, 160}, {  3,  66, 124}, {  1,  51, 100},
                    },
                    {  // band 3
                        { 89, 189, 234}, { 46, 149, 225}, { 10, 110, 194},
                        {  2,  83, 156}, {  1,  57, 113}, {  1,  47,  73},
                    },
                    {  // band 4
                        { 78, 206, 242}, { 28, 161, 232}, {  3, 114, 200},
                        {  1,  86, 161}, {  1,  62, 118}, {  1,   1,   1},
                    },
                    {  // band 5
                        { 72, 227, 250}, { 20, 182, 242}, {  3, 126, 210},
                        {  2,  91, 166}, {  1,  64, 126}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 23,  42, 227}, { 41,  43, 195}, { 25,  45, 146},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {100, 172, 245}, {165, 158, 246}, { 88, 137, 234},
                        { 44, 116, 203}, { 18,  85, 149}, {  7,  56,  92},
                    },
                    {  // band 2
                        {117, 188, 247}, { 70, 155, 239}, { 18, 105, 204},
                        {  7,  78, 158}, {  2,  50, 111}, {  1,  38,  77},
                    },
                    {  // band 3
                        {104, 207, 250}, { 54, 166, 241}, {  6, 110, 199},
                        {  1,  78, 155}, {  1,  45, 100}, {  1,   1,   1},
                    },
                    {  // band 4
                        { 87, 216, 251}, { 30, 177, 243}, {  1, 114, 203},
                        {  1,  85, 157}, {  1,  53, 108}, {128, 128, 128},
                    },
                    {  // band 5
                        { 80, 230, 253}, { 23, 193, 248}, {  1, 127, 215},
                        {  1,  94, 170}, {  1,  71,  59}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {222,   9, 234}, {161,  20, 210}, {113,  30, 185},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {195, 120, 248}, {231, 124, 247}, {148, 116, 238},
                        { 64,  98, 207}, { 20,  70, 147}, { 87,  68, 100},
                    },
                    {  // band 2
                        {186, 161, 250}, {124, 148, 245}, { 44, 123, 230},
                        { 23, 107, 205}, {  1,  80, 131}, {128, 128, 128},
                    },
                    {  // band 3
                        {172, 196, 252}, {110, 160, 248}, { 37, 134, 235},
                        { 23, 125, 200}, {128, 254, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {173, 209, 253}, {103, 175, 250}, {  1, 120, 240},
                        {  1, 146, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {184, 235, 254}, { 81, 186, 251}, {128, 109, 254},
                        {128, 254, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {248,   8, 243}, {185,  11, 225}, {108,  11, 189},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {208, 158, 254}, {244, 147, 252}, {195, 132, 248},
                        {161, 122, 224}, {129, 114, 188}, { 59, 119, 159},
                    },
                    {  // band 2
                        {202, 182, 253}, {143, 161, 251}, { 73, 115, 247},
                        {146, 175, 204}, {128,   1, 254}, {128, 128, 128},
                    },
                    {  // band 3
                        {202, 204, 254}, {131, 174, 251}, { 18, 153, 207},
                        {128, 254, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {192, 221, 254}, {114, 190, 254}, {128, 170, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {166, 236, 254}, {119, 200, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 30,  32, 144}, { 21,  35,  96}, {  4,  27,  55},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35, 107, 172}, { 61, 104, 170}, { 33,  94, 160},
                        { 13,  80, 139}, {  2,  55,  97}, {  1,  28,  49},
                    },
                    {  // band 2
                        { 51, 153, 195}, { 29, 129, 189}, {  9,  99, 163},
                        {  3,  75, 129}, {  1,  49,  88}, {  1,  29,  50},
                    },
                    {  // band 3
                        { 53, 164, 210}, { 21, 134, 201}, {  3,  97, 164},
                        {  1,  69, 124}, {  1,  45,  82}, {  1,  31,  58},
                    },
                    {  // band 4
                        { 47, 205, 234}, { 18, 158, 220}, {  2, 109, 177},
                        {  1,  78, 137}, {  1,  53, 101}, {  1,  34,  70},
                    },
                    {  // band 5
                        { 55, 233, 245}, { 16, 179, 233}, {  1, 116, 191},
                        {  1,  79, 145}, {  1,  53, 101}, {  1,  37,  58},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 36,  33, 227}, { 39,  28, 190}, { 18,  27, 134},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 76, 156, 235}, {184, 147, 235}, {114, 130, 220},
                        { 72, 112, 191}, { 42,  87, 144}, { 21,  65,  93},
                    },
                    {  // band 2
                        { 96, 179, 240}, { 51, 149, 228}, { 12, 105, 191},
                        {  6,  74, 148}, {  1,  47, 100}, {  1,  29,  53},
                    },
                    {  // band 3
                        { 88, 191, 242}, { 35, 154, 231}, {  3, 106, 187},
                        {  1,  74, 140}, {  1,  41,  84}, {  1,  25,  38},
                    },
                    {  // band 4
                        { 77, 212, 249}, { 28, 171, 239}, {  2, 117, 199},
                        {  1,  79, 151}, {  1,  45,  99}, {  1,   1,   1},
                    },
                    {  // band 5
                        { 77, 236, 252}, { 27, 190, 246}, {  2, 120, 203},
                        {  1,  78, 147}, {  1,  42,  72}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {185,  11, 227}, {113,  30, 182}, { 57,  44, 144},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {151, 139, 244}, {212, 139, 241}, {124, 126, 231},
                        { 59, 104, 213}, { 26,  73, 158}, { 20,  45,  95},
                    },
                    {  // band 2
                        {155, 163, 247}, {108, 152, 239}, { 39, 124, 214},
                        {  7, 109, 162}, { 29,  57, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {158, 176, 250}, { 89, 164, 243}, { 11, 114, 196},
                        {  1,  96, 141}, {  1,  81, 118}, {128,   1,   1},
                    },
                    {  // band 4
                        {148, 212, 251}, { 59, 174, 240}, {  2, 130, 203},
                        {  1,  70, 168}, {  1,  51, 106}, {128, 128, 128},
                    },
                    {  // band 5
                        {104, 237, 252}, { 39, 190, 246}, {  1, 154, 220},
                        {128, 102,   1}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {236,   6, 242}, {111,   6, 206}, { 36,   5, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {193, 193, 252}, {248, 182, 251}, {218, 150, 246},
                        {182, 134, 244}, {151, 137, 227}, { 45, 102, 195},
                    },
                    {  // band 2
                        {188, 202, 251}, {125, 165, 249}, { 64,  75, 218},
                        {  1, 128, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {178, 225, 254}, {107, 188, 231}, { 21, 135, 233},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {164, 227, 253}, { 55, 193, 251}, {  1, 111, 225},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {151, 243, 254}, { 50, 203, 254}, {128, 179, 254},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
    },
};
#else
static const av1_coeff_probs_model default_coef_probs_4x4[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        { 195,  29, 183 }, {  84,  49, 136 }, {   8,  42,  71 }
      }, {  // Band 1
        {  31, 107, 169 }, {  35,  99, 159 }, {  17,  82, 140 },
        {   8,  66, 114 }, {   2,  44,  76 }, {   1,  19,  32 }
      }, {  // Band 2
        {  40, 132, 201 }, {  29, 114, 187 }, {  13,  91, 157 },
        {   7,  75, 127 }, {   3,  58,  95 }, {   1,  28,  47 }
      }, {  // Band 3
        {  69, 142, 221 }, {  42, 122, 201 }, {  15,  91, 159 },
        {   6,  67, 121 }, {   1,  42,  77 }, {   1,  17,  31 }
      }, {  // Band 4
        { 102, 148, 228 }, {  67, 117, 204 }, {  17,  82, 154 },
        {   6,  59, 114 }, {   2,  39,  75 }, {   1,  15,  29 }
      }, {  // Band 5
        { 156,  57, 233 }, { 119,  57, 212 }, {  58,  48, 163 },
        {  29,  40, 124 }, {  12,  30,  81 }, {   3,  12,  31 }
      }
    }, {  // Inter
      {  // Band 0
        { 191, 107, 226 }, { 124, 117, 204 }, {  25,  99, 155 }
      }, {  // Band 1
        {  29, 148, 210 }, {  37, 126, 194 }, {   8,  93, 157 },
        {   2,  68, 118 }, {   1,  39,  69 }, {   1,  17,  33 }
      }, {  // Band 2
        {  41, 151, 213 }, {  27, 123, 193 }, {   3,  82, 144 },
        {   1,  58, 105 }, {   1,  32,  60 }, {   1,  13,  26 }
      }, {  // Band 3
        {  59, 159, 220 }, {  23, 126, 198 }, {   4,  88, 151 },
        {   1,  66, 114 }, {   1,  38,  71 }, {   1,  18,  34 }
      }, {  // Band 4
        { 114, 136, 232 }, {  51, 114, 207 }, {  11,  83, 155 },
        {   3,  56, 105 }, {   1,  33,  65 }, {   1,  17,  34 }
      }, {  // Band 5
        { 149,  65, 234 }, { 121,  57, 215 }, {  61,  49, 166 },
        {  28,  36, 114 }, {  12,  25,  76 }, {   3,  16,  42 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 214,  49, 220 }, { 132,  63, 188 }, {  42,  65, 137 }
      }, {  // Band 1
        {  85, 137, 221 }, { 104, 131, 216 }, {  49, 111, 192 },
        {  21,  87, 155 }, {   2,  49,  87 }, {   1,  16,  28 }
      }, {  // Band 2
        {  89, 163, 230 }, {  90, 137, 220 }, {  29, 100, 183 },
        {  10,  70, 135 }, {   2,  42,  81 }, {   1,  17,  33 }
      }, {  // Band 3
        { 108, 167, 237 }, {  55, 133, 222 }, {  15,  97, 179 },
        {   4,  72, 135 }, {   1,  45,  85 }, {   1,  19,  38 }
      }, {  // Band 4
        { 124, 146, 240 }, {  66, 124, 224 }, {  17,  88, 175 },
        {   4,  58, 122 }, {   1,  36,  75 }, {   1,  18,  37 }
      }, {  //  Band 5
        { 141,  79, 241 }, { 126,  70, 227 }, {  66,  58, 182 },
        {  30,  44, 136 }, {  12,  34,  96 }, {   2,  20,  47 }
      }
    }, {  // Inter
      {  // Band 0
        { 229,  99, 249 }, { 143, 111, 235 }, {  46, 109, 192 }
      }, {  // Band 1
        {  82, 158, 236 }, {  94, 146, 224 }, {  25, 117, 191 },
        {   9,  87, 149 }, {   3,  56,  99 }, {   1,  33,  57 }
      }, {  // Band 2
        {  83, 167, 237 }, {  68, 145, 222 }, {  10, 103, 177 },
        {   2,  72, 131 }, {   1,  41,  79 }, {   1,  20,  39 }
      }, {  // Band 3
        {  99, 167, 239 }, {  47, 141, 224 }, {  10, 104, 178 },
        {   2,  73, 133 }, {   1,  44,  85 }, {   1,  22,  47 }
      }, {  // Band 4
        { 127, 145, 243 }, {  71, 129, 228 }, {  17,  93, 177 },
        {   3,  61, 124 }, {   1,  41,  84 }, {   1,  21,  52 }
      }, {  // Band 5
        { 157,  78, 244 }, { 140,  72, 231 }, {  69,  58, 184 },
        {  31,  44, 137 }, {  14,  38, 105 }, {   8,  23,  61 }
      }
    }
  }
};

static const av1_coeff_probs_model default_coef_probs_8x8[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        { 125,  34, 187 }, {  52,  41, 133 }, {   6,  31,  56 }
      }, {  // Band 1
        {  37, 109, 153 }, {  51, 102, 147 }, {  23,  87, 128 },
        {   8,  67, 101 }, {   1,  41,  63 }, {   1,  19,  29 }
      }, {  // Band 2
        {  31, 154, 185 }, {  17, 127, 175 }, {   6,  96, 145 },
        {   2,  73, 114 }, {   1,  51,  82 }, {   1,  28,  45 }
      }, {  // Band 3
        {  23, 163, 200 }, {  10, 131, 185 }, {   2,  93, 148 },
        {   1,  67, 111 }, {   1,  41,  69 }, {   1,  14,  24 }
      }, {  // Band 4
        {  29, 176, 217 }, {  12, 145, 201 }, {   3, 101, 156 },
        {   1,  69, 111 }, {   1,  39,  63 }, {   1,  14,  23 }
      }, {  // Band 5
        {  57, 192, 233 }, {  25, 154, 215 }, {   6, 109, 167 },
        {   3,  78, 118 }, {   1,  48,  69 }, {   1,  21,  29 }
      }
    }, {  // Inter
      {  // Band 0
        { 202, 105, 245 }, { 108, 106, 216 }, {  18,  90, 144 }
      }, {  // Band 1
        {  33, 172, 219 }, {  64, 149, 206 }, {  14, 117, 177 },
        {   5,  90, 141 }, {   2,  61,  95 }, {   1,  37,  57 }
      }, {  // Band 2
        {  33, 179, 220 }, {  11, 140, 198 }, {   1,  89, 148 },
        {   1,  60, 104 }, {   1,  33,  57 }, {   1,  12,  21 }
      }, {  // Band 3
        {  30, 181, 221 }, {   8, 141, 198 }, {   1,  87, 145 },
        {   1,  58, 100 }, {   1,  31,  55 }, {   1,  12,  20 }
      }, {  // Band 4
        {  32, 186, 224 }, {   7, 142, 198 }, {   1,  86, 143 },
        {   1,  58, 100 }, {   1,  31,  55 }, {   1,  12,  22 }
      }, {  // Band 5
        {  57, 192, 227 }, {  20, 143, 204 }, {   3,  96, 154 },
        {   1,  68, 112 }, {   1,  42,  69 }, {   1,  19,  32 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 212,  35, 215 }, { 113,  47, 169 }, {  29,  48, 105 }
      }, {  // Band 1
        {  74, 129, 203 }, { 106, 120, 203 }, {  49, 107, 178 },
        {  19,  84, 144 }, {   4,  50,  84 }, {   1,  15,  25 }
      }, {  // Band 2
        {  71, 172, 217 }, {  44, 141, 209 }, {  15, 102, 173 },
        {   6,  76, 133 }, {   2,  51,  89 }, {   1,  24,  42 }
      }, {  // Band 3
        {  64, 185, 231 }, {  31, 148, 216 }, {   8, 103, 175 },
        {   3,  74, 131 }, {   1,  46,  81 }, {   1,  18,  30 }
      }, {  // Band 4
        {  65, 196, 235 }, {  25, 157, 221 }, {   5, 105, 174 },
        {   1,  67, 120 }, {   1,  38,  69 }, {   1,  15,  30 }
      }, {  // Band 5
        {  65, 204, 238 }, {  30, 156, 224 }, {   7, 107, 177 },
        {   2,  70, 124 }, {   1,  42,  73 }, {   1,  18,  34 }
      }
    }, {  // Inter
      {  // Band 0
        { 225,  86, 251 }, { 144, 104, 235 }, {  42,  99, 181 }
      }, {  // Band 1
        {  85, 175, 239 }, { 112, 165, 229 }, {  29, 136, 200 },
        {  12, 103, 162 }, {   6,  77, 123 }, {   2,  53,  84 }
      }, {  // Band 2
        {  75, 183, 239 }, {  30, 155, 221 }, {   3, 106, 171 },
        {   1,  74, 128 }, {   1,  44,  76 }, {   1,  17,  28 }
      }, {  // Band 3
        {  73, 185, 240 }, {  27, 159, 222 }, {   2, 107, 172 },
        {   1,  75, 127 }, {   1,  42,  73 }, {   1,  17,  29 }
      }, {  // Band 4
        {  62, 190, 238 }, {  21, 159, 222 }, {   2, 107, 172 },
        {   1,  72, 122 }, {   1,  40,  71 }, {   1,  18,  32 }
      }, {  // Band 5
        {  61, 199, 240 }, {  27, 161, 226 }, {   4, 113, 180 },
        {   1,  76, 129 }, {   1,  46,  80 }, {   1,  23,  41 }
      }
    }
  }
};

static const av1_coeff_probs_model default_coef_probs_16x16[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        {   7,  27, 153 }, {   5,  30,  95 }, {   1,  16,  30 }
      }, {  // Band 1
        {  50,  75, 127 }, {  57,  75, 124 }, {  27,  67, 108 },
        {  10,  54,  86 }, {   1,  33,  52 }, {   1,  12,  18 }
      }, {  // Band 2
        {  43, 125, 151 }, {  26, 108, 148 }, {   7,  83, 122 },
        {   2,  59,  89 }, {   1,  38,  60 }, {   1,  17,  27 }
      }, {  // Band 3
        {  23, 144, 163 }, {  13, 112, 154 }, {   2,  75, 117 },
        {   1,  50,  81 }, {   1,  31,  51 }, {   1,  14,  23 }
      }, {  // Band 4
        {  18, 162, 185 }, {   6, 123, 171 }, {   1,  78, 125 },
        {   1,  51,  86 }, {   1,  31,  54 }, {   1,  14,  23 }
      }, {  // Band 5
        {  15, 199, 227 }, {   3, 150, 204 }, {   1,  91, 146 },
        {   1,  55,  95 }, {   1,  30,  53 }, {   1,  11,  20 }
      }
    }, {  // Inter
      {  // Band 0
        {  19,  55, 240 }, {  19,  59, 196 }, {   3,  52, 105 }
      }, {  // Band 1
        {  41, 166, 207 }, { 104, 153, 199 }, {  31, 123, 181 },
        {  14, 101, 152 }, {   5,  72, 106 }, {   1,  36,  52 }
      }, {  // Band 2
        {  35, 176, 211 }, {  12, 131, 190 }, {   2,  88, 144 },
        {   1,  60, 101 }, {   1,  36,  60 }, {   1,  16,  28 }
      }, {  // Band 3
        {  28, 183, 213 }, {   8, 134, 191 }, {   1,  86, 142 },
        {   1,  56,  96 }, {   1,  30,  53 }, {   1,  12,  20 }
      }, {  // Band 4
        {  20, 190, 215 }, {   4, 135, 192 }, {   1,  84, 139 },
        {   1,  53,  91 }, {   1,  28,  49 }, {   1,  11,  20 }
      }, {  // Band 5
        {  13, 196, 216 }, {   2, 137, 192 }, {   1,  86, 143 },
        {   1,  57,  99 }, {   1,  32,  56 }, {   1,  13,  24 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 211,  29, 217 }, {  96,  47, 156 }, {  22,  43,  87 }
      }, {  // Band 1
        {  78, 120, 193 }, { 111, 116, 186 }, {  46, 102, 164 },
        {  15,  80, 128 }, {   2,  49,  76 }, {   1,  18,  28 }
      }, {  // Band 2
        {  71, 161, 203 }, {  42, 132, 192 }, {  10,  98, 150 },
        {   3,  69, 109 }, {   1,  44,  70 }, {   1,  18,  29 }
      }, {  // Band 3
        {  57, 186, 211 }, {  30, 140, 196 }, {   4,  93, 146 },
        {   1,  62, 102 }, {   1,  38,  65 }, {   1,  16,  27 }
      }, {  // Band 4
        {  47, 199, 217 }, {  14, 145, 196 }, {   1,  88, 142 },
        {   1,  57,  98 }, {   1,  36,  62 }, {   1,  15,  26 }
      }, {  // Band 5
        {  26, 219, 229 }, {   5, 155, 207 }, {   1,  94, 151 },
        {   1,  60, 104 }, {   1,  36,  62 }, {   1,  16,  28 }
      }
    }, {  // Inter
      {  // Band 0
        { 233,  29, 248 }, { 146,  47, 220 }, {  43,  52, 140 }
      }, {  // Band 1
        { 100, 163, 232 }, { 179, 161, 222 }, {  63, 142, 204 },
        {  37, 113, 174 }, {  26,  89, 137 }, {  18,  68,  97 }
      }, {  // Band 2
        {  85, 181, 230 }, {  32, 146, 209 }, {   7, 100, 164 },
        {   3,  71, 121 }, {   1,  45,  77 }, {   1,  18,  30 }
      }, {  // Band 3
        {  65, 187, 230 }, {  20, 148, 207 }, {   2,  97, 159 },
        {   1,  68, 116 }, {   1,  40,  70 }, {   1,  14,  29 }
      }, {  // Band 4
        {  40, 194, 227 }, {   8, 147, 204 }, {   1,  94, 155 },
        {   1,  65, 112 }, {   1,  39,  66 }, {   1,  14,  26 }
      }, {  // Band 5
        {  16, 208, 228 }, {   3, 151, 207 }, {   1,  98, 160 },
        {   1,  67, 117 }, {   1,  41,  74 }, {   1,  17,  31 }
      }
    }
  }
};

static const av1_coeff_probs_model default_coef_probs_32x32[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        {  17,  38, 140 }, {   7,  34,  80 }, {   1,  17,  29 }
      }, {  // Band 1
        {  37,  75, 128 }, {  41,  76, 128 }, {  26,  66, 116 },
        {  12,  52,  94 }, {   2,  32,  55 }, {   1,  10,  16 }
      }, {  // Band 2
        {  50, 127, 154 }, {  37, 109, 152 }, {  16,  82, 121 },
        {   5,  59,  85 }, {   1,  35,  54 }, {   1,  13,  20 }
      }, {  // Band 3
        {  40, 142, 167 }, {  17, 110, 157 }, {   2,  71, 112 },
        {   1,  44,  72 }, {   1,  27,  45 }, {   1,  11,  17 }
      }, {  // Band 4
        {  30, 175, 188 }, {   9, 124, 169 }, {   1,  74, 116 },
        {   1,  48,  78 }, {   1,  30,  49 }, {   1,  11,  18 }
      }, {  // Band 5
        {  10, 222, 223 }, {   2, 150, 194 }, {   1,  83, 128 },
        {   1,  48,  79 }, {   1,  27,  45 }, {   1,  11,  17 }
      }
    }, {  // Inter
      {  // Band 0
        {  36,  41, 235 }, {  29,  36, 193 }, {  10,  27, 111 }
      }, {  // Band 1
        {  85, 165, 222 }, { 177, 162, 215 }, { 110, 135, 195 },
        {  57, 113, 168 }, {  23,  83, 120 }, {  10,  49,  61 }
      }, {  // Band 2
        {  85, 190, 223 }, {  36, 139, 200 }, {   5,  90, 146 },
        {   1,  60, 103 }, {   1,  38,  65 }, {   1,  18,  30 }
      }, {  // Band 3
        {  72, 202, 223 }, {  23, 141, 199 }, {   2,  86, 140 },
        {   1,  56,  97 }, {   1,  36,  61 }, {   1,  16,  27 }
      }, {  // Band 4
        {  55, 218, 225 }, {  13, 145, 200 }, {   1,  86, 141 },
        {   1,  57,  99 }, {   1,  35,  61 }, {   1,  13,  22 }
      }, {  // Band 5
        {  15, 235, 212 }, {   1, 132, 184 }, {   1,  84, 139 },
        {   1,  57,  97 }, {   1,  34,  56 }, {   1,  14,  23 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 181,  21, 201 }, {  61,  37, 123 }, {  10,  38,  71 }
      }, {  // Band 1
        {  47, 106, 172 }, {  95, 104, 173 }, {  42,  93, 159 },
        {  18,  77, 131 }, {   4,  50,  81 }, {   1,  17,  23 }
      }, {  // Band 2
        {  62, 147, 199 }, {  44, 130, 189 }, {  28, 102, 154 },
        {  18,  75, 115 }, {   2,  44,  65 }, {   1,  12,  19 }
      }, {  // Band 3
        {  55, 153, 210 }, {  24, 130, 194 }, {   3,  93, 146 },
        {   1,  61,  97 }, {   1,  31,  50 }, {   1,  10,  16 }
      }, {  // Band 4
        {  49, 186, 223 }, {  17, 148, 204 }, {   1,  96, 142 },
        {   1,  53,  83 }, {   1,  26,  44 }, {   1,  11,  17 }
      }, {  // Band 5
        {  13, 217, 212 }, {   2, 136, 180 }, {   1,  78, 124 },
        {   1,  50,  83 }, {   1,  29,  49 }, {   1,  14,  23 }
      }
    }, {  // Inter
      {  // Band 0
        { 197,  13, 247 }, {  82,  17, 222 }, {  25,  17, 162 }
      }, {  // Band 1
        { 126, 186, 247 }, { 234, 191, 243 }, { 176, 177, 234 },
        { 104, 158, 220 }, {  66, 128, 186 }, {  55,  90, 137 }
      }, {  // Band 2
        { 111, 197, 242 }, {  46, 158, 219 }, {   9, 104, 171 },
        {   2,  65, 125 }, {   1,  44,  80 }, {   1,  17,  91 }
      }, {  // Band 3
        { 104, 208, 245 }, {  39, 168, 224 }, {   3, 109, 162 },
        {   1,  79, 124 }, {   1,  50, 102 }, {   1,  43, 102 }
      }, {  // Band 4
        {  84, 220, 246 }, {  31, 177, 231 }, {   2, 115, 180 },
        {   1,  79, 134 }, {   1,  55,  77 }, {   1,  60,  79 }
      }, {  // Band 5
        {  43, 243, 240 }, {   8, 180, 217 }, {   1, 115, 166 },
        {   1,  84, 121 }, {   1,  51,  67 }, {   1,  16,   6 }
      }
    }
  }
};
#endif  // CONFIG_ENTROPY
/* clang-format on */

static void extend_to_full_distribution(aom_prob *probs, aom_prob p) {
  assert(p != 0);
  memcpy(probs, av1_pareto8_full[p - 1], MODEL_NODES * sizeof(aom_prob));
}

void av1_model_to_full_probs(const aom_prob *model, aom_prob *full) {
  if (full != model)
    memcpy(full, model, sizeof(aom_prob) * UNCONSTRAINED_NODES);
  extend_to_full_distribution(&full[UNCONSTRAINED_NODES], model[PIVOT_NODE]);
}

#if CONFIG_ANS
void av1_build_token_cdfs(const aom_prob *pdf_model, rans_dec_lut cdf) {
  AnsP10 pdf_tab[ENTROPY_TOKENS - 1];
  assert(pdf_model[2] != 0);
  // TODO(aconverse): Investigate making the precision of the zero and EOB tree
  // nodes 10-bits.
  aom_rans_merge_prob8_pdf(pdf_tab, pdf_model[1],
                           av1_pareto8_token_probs[pdf_model[2] - 1],
                           ENTROPY_TOKENS - 2);
  aom_rans_build_cdf_from_pdf(pdf_tab, cdf);
}

void av1_coef_pareto_cdfs(FRAME_CONTEXT *fc) {
  TX_SIZE t;
  int i, j, k, l;
  for (t = TX_4X4; t <= TX_32X32; ++t)
    for (i = 0; i < PLANE_TYPES; ++i)
      for (j = 0; j < REF_TYPES; ++j)
        for (k = 0; k < COEF_BANDS; ++k)
          for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l)
            av1_build_token_cdfs(fc->coef_probs[t][i][j][k][l],
                                 fc->coef_cdfs[t][i][j][k][l]);
}
#endif  // CONFIG_ANS

void av1_default_coef_probs(AV1_COMMON *cm) {
#if CONFIG_ENTROPY
  const int index = AOMMIN(
      ROUND_POWER_OF_TWO(cm->base_qindex, 8 - QCTX_BIN_BITS), QCTX_BINS - 1);
  av1_copy(cm->fc->coef_probs, default_qctx_coef_probs[index]);
#else
  av1_copy(cm->fc->coef_probs[TX_4X4], default_coef_probs_4x4);
  av1_copy(cm->fc->coef_probs[TX_8X8], default_coef_probs_8x8);
  av1_copy(cm->fc->coef_probs[TX_16X16], default_coef_probs_16x16);
  av1_copy(cm->fc->coef_probs[TX_32X32], default_coef_probs_32x32);
#endif  // CONFIG_ENTROPY
#if CONFIG_ANS
  av1_coef_pareto_cdfs(cm->fc);
#endif  // CONFIG_ANS
}

static void adapt_coef_probs(AV1_COMMON *cm, TX_SIZE tx_size,
                             unsigned int count_sat,
                             unsigned int update_factor) {
  const FRAME_CONTEXT *pre_fc = &cm->frame_contexts[cm->frame_context_idx];
  av1_coeff_probs_model *const probs = cm->fc->coef_probs[tx_size];
#if CONFIG_ENTROPY
  const av1_coeff_probs_model *const pre_probs =
      cm->partial_prob_update
          ? (const av1_coeff_probs_model *)cm->starting_coef_probs[tx_size]
          : pre_fc->coef_probs[tx_size];
#else
  const av1_coeff_probs_model *const pre_probs = pre_fc->coef_probs[tx_size];
#endif  // CONFIG_ENTROPY
  const av1_coeff_count_model *const counts =
      (const av1_coeff_count_model *)cm->counts.coef[tx_size];
  const unsigned int(*eob_counts)[REF_TYPES][COEF_BANDS][COEFF_CONTEXTS] =
      (const unsigned int(*)[REF_TYPES][COEF_BANDS]
                            [COEFF_CONTEXTS])cm->counts.eob_branch[tx_size];
  int i, j, k, l, m;

  for (i = 0; i < PLANE_TYPES; ++i)
    for (j = 0; j < REF_TYPES; ++j)
      for (k = 0; k < COEF_BANDS; ++k)
        for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l) {
          const int n0 = counts[i][j][k][l][ZERO_TOKEN];
          const int n1 = counts[i][j][k][l][ONE_TOKEN];
          const int n2 = counts[i][j][k][l][TWO_TOKEN];
          const int neob = counts[i][j][k][l][EOB_MODEL_TOKEN];
          const unsigned int branch_ct[UNCONSTRAINED_NODES][2] = {
            { neob, eob_counts[i][j][k][l] - neob }, { n0, n1 + n2 }, { n1, n2 }
          };
          for (m = 0; m < UNCONSTRAINED_NODES; ++m)
            probs[i][j][k][l][m] =
                av1_merge_probs(pre_probs[i][j][k][l][m], branch_ct[m],
                                count_sat, update_factor);
        }
}

void av1_adapt_coef_probs(AV1_COMMON *cm) {
  TX_SIZE t;
  unsigned int count_sat, update_factor;

#if CONFIG_ENTROPY
  if (cm->last_frame_type == KEY_FRAME) {
    update_factor = COEF_MAX_UPDATE_FACTOR_AFTER_KEY_BITS; /* adapt quickly */
    count_sat = COEF_COUNT_SAT_AFTER_KEY_BITS;
  } else {
    update_factor = COEF_MAX_UPDATE_FACTOR_BITS;
    count_sat = COEF_COUNT_SAT_BITS;
  }
  if (cm->partial_prob_update == 1) {
    update_factor = COEF_MAX_UPDATE_FACTOR_BITS;
  }
#else
  if (cm->last_frame_type == KEY_FRAME) {
    update_factor = COEF_MAX_UPDATE_FACTOR_AFTER_KEY; /* adapt quickly */
    count_sat = COEF_COUNT_SAT_AFTER_KEY;
  } else {
    update_factor = COEF_MAX_UPDATE_FACTOR;
    count_sat = COEF_COUNT_SAT;
  }
#endif  // CONFIG_ENTROPY
  for (t = TX_4X4; t <= TX_32X32; t++)
    adapt_coef_probs(cm, t, count_sat, update_factor);
#if CONFIG_ANS
  av1_coef_pareto_cdfs(cm->fc);
#endif
}

#if CONFIG_ENTROPY
void av1_partial_adapt_probs(AV1_COMMON *cm, int mi_row, int mi_col) {
  (void)mi_row;
  (void)mi_col;

  if (cm->refresh_frame_context == REFRESH_FRAME_CONTEXT_BACKWARD) {
    cm->partial_prob_update = 1;
    av1_adapt_coef_probs(cm);
  }
}
#endif  // CONFIG_ENTROPY
